import math
import statistics

import numpy as np
import pytest

from hstrat import hstrat
import hstrat.phylogenetic_inference.priors._detail as detail


def test_base_class():
    assert issubclass(hstrat.ArbitraryPrior, detail.PriorBase)


def test_calc_interval_probability_proxy():
    prior = hstrat.ArbitraryPrior()

    assert (
        len(
            {
                prior.CalcIntervalProbabilityProxy(0, 50),
                prior.CalcIntervalProbabilityProxy(7, 8),
                prior.CalcIntervalProbabilityProxy(99, 100),
                prior.CalcIntervalProbabilityProxy(78, 100),
            }
        )
        == 1
    )


@pytest.mark.parametrize(
    "interval_width",
    [1, 2, 3],
)
def test_calc_interval_conditioned_mean(interval_width):
    prior = hstrat.ArbitraryPrior()
    begin, end = 0, 10000

    # required for arbitrary prior due to arbitraryness
    end -= end % interval_width

    intervals = [
        (interval_begin, interval_begin + interval_width)
        for interval_begin in range(begin, end, interval_width)
    ]

    samples = np.array(
        [
            statistics.mean((interval_begin, interval_end - 1))
            for interval_begin, interval_end in intervals
        ]
    )
    weights = np.array(
        [
            prior.CalcIntervalProbabilityProxy(interval_begin, interval_end)
            for interval_begin, interval_end in intervals
        ]
    )

    assert math.isclose(
        np.average(samples, weights=weights),
        prior.CalcIntervalConditionedMean(begin, end),
    )


def test_sample_interval_conditioned_value():
    prior = hstrat.ArbitraryPrior()
    samples = set(
        prior.SampleIntervalConditionedValue(0, 100) for _ in range(100)
    )
    assert len(samples) > 1
    assert all(0 <= sample < 100 for sample in samples)



import typing

from .._detail._PolicyCouplerBase import PolicyCouplerBase
from ._CalcMrcaUncertaintyAbsUpperBoundWorstCase import (
    CalcMrcaUncertaintyAbsUpperBoundWorstCase,
)


class CalcMrcaUncertaintyRelUpperBoundWorstCase:
    def __init__(
        self: "CalcMrcaUncertaintyRelUpperBoundWorstCase",
        policy_spec: typing.Optional[typing.Any] = None,
    ) -> None:
        pass

    def __eq__(
        self: "CalcMrcaUncertaintyRelUpperBoundWorstCase",
        other: typing.Any,
    ) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcMrcaUncertaintyRelUpperBoundWorstCase",
        policy: typing.Optional[PolicyCouplerBase],
        first_num_strata_deposited: int,
        second_num_strata_deposited: int,
        actual_rank_of_mrca: int,
    ) -> float:
        """At most, how much uncertainty to relative estimate rank of MRCA?
        Inclusive."""
        # rectify negative-indexed actual_rank_of_mrca
        if actual_rank_of_mrca is not None and actual_rank_of_mrca < 0:
            least_last_rank = min(
                first_num_strata_deposited - 1,
                second_num_strata_deposited - 1,
            )
            actual_rank_of_mrca += least_last_rank
            assert actual_rank_of_mrca >= 0

        if 0 in (first_num_strata_deposited, second_num_strata_deposited):
            return 0.0

        least_last_rank = min(
            first_num_strata_deposited - 1,
            second_num_strata_deposited - 1,
        )

        # conservatively normalize by smallest ranks since mrca
        min_ranks_since_mrca = least_last_rank - actual_rank_of_mrca
        worst_abs_uncertainty = CalcMrcaUncertaintyAbsUpperBoundWorstCase()(
            policy,
            first_num_strata_deposited,
            second_num_strata_deposited,
            actual_rank_of_mrca,
        )

        if min_ranks_since_mrca == 0:
            return 0.0
        else:
            return worst_abs_uncertainty / min_ranks_since_mrca



import itertools as it

import pytest

from hstrat._auxiliary_lib import random_choice_generator


def test_returns_one_value():
    # Arrange
    values = [1, 2, 3]
    generator = random_choice_generator(values)

    # Act
    result = next(generator)

    # Assert
    assert result in values
    assert isinstance(result, int)


def test_returns_n_values():
    # Arrange
    values = [1, 2, 3]
    n = 5
    generator = random_choice_generator(values)

    # Act
    results = [next(generator) for _ in range(n)]

    # Assert
    assert all(result in values for result in results)
    assert all(isinstance(result, int) for result in results)
    assert set(values) == set(it.islice(generator, 0, 100))


def test_raises_exception_for_empty_list():
    # Arrange
    values = []
    generator = random_choice_generator(values)

    # Act / Assert
    with pytest.raises(IndexError):
        next(generator)


def test_negative_values():
    # Arrange
    values = [-1, 0, 1]
    generator = random_choice_generator(values)

    # Act / Assert
    for val in it.islice(generator, 0, 100):
        assert val in values

    assert set(it.islice(generator, 0, 100)) == set(values)


def test_mixed_values():
    # Arrange
    values = ["string", None, 3, 4.5]
    generator = random_choice_generator(values)

    # Act / Assert
    for val in it.islice(generator, 0, 100):
        assert val in values

    assert set(it.islice(generator, 0, 100)) == set(values)



import typing

from .._detail._PolicyCouplerBase import PolicyCouplerBase


class CalcMrcaUncertaintyRelExactFromAbs:
    def __init__(
        self: "CalcMrcaUncertaintyRelExactFromAbs",
        policy_spec: typing.Optional[typing.Any] = None,
    ) -> None:
        pass

    def __eq__(
        self: "CalcMrcaUncertaintyRelExactFromAbs",
        other: typing.Any,
    ) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcMrcaUncertaintyRelExactFromAbs",
        policy: typing.Optional[PolicyCouplerBase],
        first_num_strata_deposited: int,
        second_num_strata_deposited: int,
        actual_rank_of_mrca: int,
    ) -> float:
        """At most, how much absolute uncertainty to estimate rank of MRCA?
        Inclusive."""
        # rectify negative-indexed actual_rank_of_mrca
        if actual_rank_of_mrca is not None and actual_rank_of_mrca < 0:
            least_last_rank = min(
                first_num_strata_deposited - 1,
                second_num_strata_deposited - 1,
            )
            actual_rank_of_mrca += least_last_rank
            assert actual_rank_of_mrca >= 0

        if (
            first_num_strata_deposited <= 2
            or second_num_strata_deposited <= 2
            or actual_rank_of_mrca
            in (
                first_num_strata_deposited - 1,
                second_num_strata_deposited - 1,
            )
        ):
            return 0.0

        uncertainty_abs_exact = policy.CalcMrcaUncertaintyAbsExact(
            first_num_strata_deposited,
            second_num_strata_deposited,
            actual_rank_of_mrca,
        )

        least_last_rank = min(
            first_num_strata_deposited - 1,
            second_num_strata_deposited - 1,
        )
        least_recency = least_last_rank - actual_rank_of_mrca

        return uncertainty_abs_exact / least_recency



import typing


def reversed_range(
    start: int, stop: typing.Optional[int] = None
) -> typing.Iterator[int]:
    """Iterate through a range in reverse order.

    Equivalent to `reversed(range(*args))`. Compatible with numba jit.
    """
    # numba jit doesn't support yield from or reversed
    cur = stop if stop is not None else start
    last = start if stop is not None else 0
    while cur > last:
        cur -= 1
        yield cur



import typing

from ....._auxiliary_lib import memoize_generator
from ..._detail import PolicyCouplerBase
from .._PolicySpec import PolicySpec
from .._impl import get_retained_ranks


class IterRetainedRanks:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "IterRetainedRanks",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __hash__(self: "IterRetainedRanks") -> int:
        """Hash object instance."""
        return 0

    def __eq__(self: "IterRetainedRanks", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    @memoize_generator()
    def __call__(
        self: "IterRetainedRanks",
        policy: PolicyCouplerBase,
        num_strata_deposited: int,
    ) -> typing.Iterator[int]:
        """Iterate over retained strata ranks at `num_strata_deposited` in
        ascending order."""
        yield from sorted(
            get_retained_ranks(
                policy,
                num_strata_deposited,
            )
        )



import unittest

from hstrat._auxiliary_lib import is_base64

base64_alphabet = """ABCDEFGHIJKLMNOPQRSTUVWXYZ\
abcdefghijklmnopqrstuvwxyz\
0123456789+/"""


class TestBase64(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_empty(self):

        assert is_base64("")
        assert is_base64(b"")

    def test_str(self):
        assert is_base64(base64_alphabet)
        assert not is_base64("*,.")

    def test_bytes(self):

        assert is_base64(base64_alphabet.encode("ascii"))
        assert is_base64(base64_alphabet.encode("utf_8"))
        assert not is_base64(b"*")

    def test_unicode(self):

        assert not is_base64("😄")
        assert not is_base64("😄".encode("utf_8"))

    def test_not_str_bytes(self):

        self.assertRaises(ValueError, is_base64, 0)
        self.assertRaises(ValueError, is_base64, list(base64_alphabet))
        self.assertRaises(ValueError, is_base64, set(base64_alphabet))


if __name__ == "__main__":
    unittest.main()



import pytest
import typing_extensions

from hstrat import genome_instrumentation, hstrat


@pytest.mark.parametrize(
    "impl",
    [genome_instrumentation.HereditaryStratigraphicColumn],
    # TODO
    # genome_instrumentation._HereditaryStratigraphicColumn_.impls,
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
        None,
    ],
)
@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "num_deposits",
    [0, 1, 6, 8, 64],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 8, 16, 32, 64],
)
def test_col_to_packet(
    impl,
    retention_policy,
    ordered_store,
    always_store_rank_in_stratum,
    num_deposits,
    differentia_bit_width,
    caplog,
):
    column = impl(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_bit_width,
        always_store_rank_in_stratum=always_store_rank_in_stratum,
    )
    for __ in range(num_deposits):
        column.DepositStratum()

    packet = hstrat.col_to_packet(column)
    assert packet == hstrat.col_to_packet(column)
    assert packet != hstrat.col_to_packet(
        column, num_strata_deposited_byte_order="little"
    )
    assert packet != hstrat.col_to_packet(
        column, num_strata_deposited_byte_width=42
    )
    assert isinstance(packet, typing_extensions.Buffer)
    assert len(packet) <= (
        (column.GetNumStrataRetained() * differentia_bit_width + 7) // 8
        + 1  # possible num padding bits header byte
        + 4
    )


@pytest.mark.parametrize(
    "impl",
    [genome_instrumentation.HereditaryStratigraphicColumn],
    # TODO
    # genome_instrumentation._HereditaryStratigraphicColumn_.impls,
)
@pytest.mark.parametrize("byte_order", ["big", "little"])
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
        None,
    ],
)
@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "num_deposits",
    [0, 1, 6, 8, 64],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 8, 16, 32, 64],
)
def test_col_to_packet_then_from_packet(
    impl,
    byte_order,
    retention_policy,
    ordered_store,
    always_store_rank_in_stratum,
    num_deposits,
    differentia_bit_width,
    caplog,
):
    column = impl(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_bit_width,
        always_store_rank_in_stratum=always_store_rank_in_stratum,
    )
    for __ in range(num_deposits):
        column.DepositStratum()

    assert hstrat.col_to_packet(column) == hstrat.col_to_packet(column)
    packet = hstrat.col_to_packet(
        column, num_strata_deposited_byte_order=byte_order
    )
    reconstituted = hstrat.col_from_packet(
        packet,
        differentia_bit_width=differentia_bit_width,
        differentiae_byte_bit_order="big",
        num_strata_deposited_byte_order=byte_order,
        stratum_retention_policy=retention_policy,
    )
    if column.GetNumStrataRetained() >= 20:
        assert reconstituted != hstrat.col_from_packet(
            packet,
            differentia_bit_width=differentia_bit_width,
            differentiae_byte_bit_order="little",
            num_strata_deposited_byte_order=byte_order,
            stratum_retention_policy=retention_policy,
        )
    assert reconstituted._ShouldOmitStratumDepositionRank()
    if (
        ordered_store == hstrat.HereditaryStratumOrderedStoreList
        and impl == genome_instrumentation.HereditaryStratigraphicColumn
        and not always_store_rank_in_stratum
    ):
        assert reconstituted == column
    else:
        target_records = hstrat.col_to_records(column)
        # always_store_rank_in_stratum setting is lost
        if "stratum_deposition_ranks" in target_records:
            del target_records["stratum_deposition_ranks"]
        assert hstrat.col_to_records(reconstituted) == target_records


@pytest.mark.parametrize(
    "impl",
    [genome_instrumentation.HereditaryStratigraphicColumn],
    # TODO
    # genome_instrumentation._HereditaryStratigraphicColumn_.impls,
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
        None,
    ],
)
@pytest.mark.parametrize("always_store_rank_in_stratum", [True, False])
@pytest.mark.parametrize(
    "num_deposits",
    [0, 1, 6, 8, 64],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 8, 16, 32, 64],
)
@pytest.mark.parametrize(
    "buffer_excess_bytes",
    [0, 1, 100],
)
@pytest.mark.parametrize("byte_order", ["big", "little"])
def test_col_to_packet_then_from_packet_buffer(
    impl,
    retention_policy,
    ordered_store,
    always_store_rank_in_stratum,
    num_deposits,
    differentia_bit_width,
    buffer_excess_bytes,
    byte_order,
    caplog,
):
    column = impl(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_bit_width,
        always_store_rank_in_stratum=always_store_rank_in_stratum,
    )
    for __ in range(num_deposits):
        column.DepositStratum()

    assert hstrat.col_to_packet(column) == hstrat.col_to_packet(column)
    packet = hstrat.col_to_packet(
        column, num_strata_deposited_byte_order=byte_order
    )
    packet_buffer = bytearray(len(packet) + buffer_excess_bytes)
    packet_buffer[: len(packet)] = packet
    reconstituted = hstrat.col_from_packet_buffer(
        packet_buffer,
        differentiae_byte_bit_order="big",
        differentia_bit_width=differentia_bit_width,
        num_strata_deposited_byte_order=byte_order,
        stratum_retention_policy=retention_policy,
    )
    if column.GetNumStrataRetained() >= 20:
        assert reconstituted != hstrat.col_from_packet_buffer(
            packet_buffer,
            differentia_bit_width=differentia_bit_width,
            differentiae_byte_bit_order="little",
            num_strata_deposited_byte_order=byte_order,
            stratum_retention_policy=retention_policy,
        )
    assert reconstituted._ShouldOmitStratumDepositionRank()
    if (
        ordered_store == hstrat.HereditaryStratumOrderedStoreList
        and impl == genome_instrumentation.HereditaryStratigraphicColumn
        and not always_store_rank_in_stratum
    ):
        assert reconstituted == column
    else:
        target_records = hstrat.col_to_records(column)
        # always_store_rank_in_stratum setting is lost
        if "stratum_deposition_ranks" in target_records:
            del target_records["stratum_deposition_ranks"]
        assert hstrat.col_to_records(reconstituted) == target_records



import math
import typing

from ..._detail import PolicyCouplerBase
from ..._impl import CalcMrcaUncertaintyAbsUpperBoundWorstCase
from .._PolicySpec import PolicySpec
from .._impl import calc_common_ratio


class CalcMrcaUncertaintyAbsUpperBound:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "CalcMrcaUncertaintyAbsUpperBound",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(
        self: "CalcMrcaUncertaintyAbsUpperBound",
        other: typing.Any,
    ) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcMrcaUncertaintyAbsUpperBound",
        policy: PolicyCouplerBase,
        first_num_strata_deposited: int,
        second_num_strata_deposited: int,
        actual_rank_of_mrca: int,
    ) -> int:
        """At most, how much relative uncertainty to estimate rank of MRCA?
        Inclusive."""
        # rectify negative-indexed actual_rank_of_mrca
        if actual_rank_of_mrca is not None and actual_rank_of_mrca < 0:
            least_last_rank = min(
                first_num_strata_deposited - 1,
                second_num_strata_deposited - 1,
            )
            actual_rank_of_mrca += least_last_rank
            assert actual_rank_of_mrca >= 0

        spec = policy.GetSpec()

        max_num_strata_deposited = max(
            first_num_strata_deposited,
            second_num_strata_deposited,
        )
        if max_num_strata_deposited == 0:
            return 0

        interspersal = spec.GetInterspersal()
        # edge case: no uncertainty guarantee for interspersal 1
        # interspersal >= 2 required for uncertainty guarantee
        if interspersal == 1:
            return max_num_strata_deposited

        max_ranks_since_mrca = max_num_strata_deposited - actual_rank_of_mrca
        # edge case: columns are identical
        if max_ranks_since_mrca == 0:
            return 0

        common_ratio = calc_common_ratio(
            spec.GetDegree(), max_num_strata_deposited
        )
        # edge case: no strata have yet been dropped
        if common_ratio == 1.0:
            return 0

        # round up to next power of common_ratio
        rounded_ranks_since_mrca = common_ratio ** int(
            math.ceil(math.log(max_ranks_since_mrca, common_ratio))
        )
        # should be leq just multiplying max_ranks_since_mrca by common_ratio
        assert (
            rounded_ranks_since_mrca <= max_ranks_since_mrca * common_ratio
            # account for representation error etc.
            or math.isclose(
                rounded_ranks_since_mrca,
                max_ranks_since_mrca * common_ratio,
            )
        )

        # account for increased resolution from interspersal
        res = int(math.ceil(rounded_ranks_since_mrca / (interspersal - 1)))

        return min(
            res,
            CalcMrcaUncertaintyAbsUpperBoundWorstCase()(
                policy,
                first_num_strata_deposited,
                second_num_strata_deposited,
                actual_rank_of_mrca,
            ),
        )



import numpy as np
import pytest

from hstrat.hstrat import nominal_resolution_algo
from hstrat.stratum_retention_strategy.stratum_retention_algorithms._impl import (
    CalcMrcaUncertaintyRelUpperBoundPessimalRankBruteForce,
)


@pytest.mark.parametrize(
    "time_sequence",
    [
        range(10**2),
        np.random.default_rng(1).integers(
            10**3,
            size=10**2,
        ),
    ],
)
def test_policy_consistency(time_sequence):
    policy = nominal_resolution_algo.Policy()
    spec = policy.GetSpec()
    instance = (
        nominal_resolution_algo.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank(
            spec,
        )
    )
    for num_strata_deposited_a in time_sequence:
        for num_strata_deposited_b in (
            num_strata_deposited_a // 3,
            num_strata_deposited_a // 2,
            num_strata_deposited_a,
            num_strata_deposited_a + 1,
            num_strata_deposited_a + 10,
            num_strata_deposited_a + 100,
        ):
            if 0 in (num_strata_deposited_a, num_strata_deposited_b):
                continue
            for actual_mrca_rank in np.random.default_rng(1).integers(
                min(num_strata_deposited_a, num_strata_deposited_b),
                size=3,
            ):
                policy_requirement = policy.CalcMrcaUncertaintyRelUpperBound(
                    num_strata_deposited_a,
                    num_strata_deposited_b,
                    CalcMrcaUncertaintyRelUpperBoundPessimalRankBruteForce()(
                        policy,
                        num_strata_deposited_a,
                        num_strata_deposited_b,
                    ),
                )
                for which in (
                    instance,
                    nominal_resolution_algo.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank(
                        spec
                    ),
                ):
                    assert (
                        which(
                            policy,
                            num_strata_deposited_a,
                            num_strata_deposited_b,
                        )
                        == policy_requirement
                    )


def test_eq():
    policy = nominal_resolution_algo.Policy()
    spec = policy.GetSpec()
    instance = (
        nominal_resolution_algo.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank(
            spec
        )
    )

    assert instance == instance
    assert (
        instance
        == nominal_resolution_algo.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank(
            spec,
        )
    )
    assert instance is not None



import typing

from ...genome_instrumentation import HereditaryStratigraphicColumn
from ._calc_rank_of_first_retained_disparity_between import (
    calc_rank_of_first_retained_disparity_between,
)
from ._calc_rank_of_last_retained_commonality_between import (
    calc_rank_of_last_retained_commonality_between,
)


def calc_rank_of_parity_segue_between(
    first: HereditaryStratigraphicColumn,
    second: HereditaryStratigraphicColumn,
    confidence_level_commonality: float,
    confidence_level_disparity: float,
) -> typing.Tuple[typing.Optional[int], typing.Optional[int]]:
    """Find last matching and first mismatching strata between columns.

    Implementation detail. Searches forward from index 0 for first mismatching
    retained differentiae between columns.
    """

    return (
        calc_rank_of_last_retained_commonality_between(
            first, second, confidence_level_commonality
        )
        if confidence_level_commonality is not None
        else None,
        calc_rank_of_first_retained_disparity_between(
            first, second, confidence_level_disparity
        )
        if confidence_level_disparity is not None
        else None,
    )



from prettytable import ALL as prettytable_ALL
from prettytable import PrettyTable

from ...genome_instrumentation import HereditaryStratigraphicColumn
from ...serialization import col_to_dataframe


def col_to_ascii(
    column: HereditaryStratigraphicColumn,
    discarded_strata: bool = True,
    key: bool = True,
    time_bookends: bool = True,
) -> str:
    """Create an ASCII table representation of HereditaryStratigraphicColumn
    state.

    Parameters
    ----------
    column : HereditaryStratigraphicColumn
        The HereditaryStratigraphicColumn object to be converted.
    discarded_strata : bool, default True
        Include discarded strata in the output?
    key : bool, default True
        Append a legend to the output?
    time_bookends : bool, default True
        Prepend and append "MOST ANCIENT" and "MOST RECENT" to the table?

    Returns
    -------
    str
        The ASCII representation of the HereditaryStratigraphicColumn object
        in tabular format.
    """
    df = col_to_dataframe(column)
    df.set_index("rank", inplace=True)

    table = PrettyTable()
    table.field_names = [
        "stratum rank",
        "stratum index",
        "stratum differentia",
    ]

    def wrap_rank(rank):
        return (
            f"({rank})"
            if column._ShouldOmitStratumDepositionRank()
            else str(rank)
        )

    for rank in (
        range(column.GetNumStrataDeposited()) if discarded_strata else df.index
    ):
        if rank in df.index:
            row = df.loc[rank]
            table.add_row(
                [
                    wrap_rank(rank),
                    row["index"],
                    f"{int(row['differentia']):x}*",
                ]
            )
        else:
            table.add_row(
                [
                    rank,
                    "".join("░" for __ in "stratum index"),
                    "".join("░" for __ in "stratum differentia"),
                ]
            )

    table.hrules = prettytable_ALL
    res = table.get_string()

    if time_bookends:
        pad = max(len(row) for row in res.split("\n")) - 2
        res = f"|{'MOST ANCIENT':^{pad}}|\n" + res
        res = "+" + "-" * pad + "+\n" + res

        res += f"\n|{'MOST RECENT':^{pad}}|"
        res += "\n+" + "-" * pad + "+"

    if key and "*" in res:
        res += "\n*: stored in stratum (otherwise calculated)"

    if key and "░" in res:
        res += "\n░: discarded stratum"

    return res



import math

from ._calc_target_recency import calc_target_recency


def calc_target_rank(
    degree: int,
    pow_: int,
    num_strata_deposited: int,
) -> int:
    """What should the rank of the `pow`th exponentially-spaced-back-from-
    recency-zero target be?

    Will monotonically increase with `num_strata_deposited`.
    """
    target_recency = calc_target_recency(degree, pow_, num_strata_deposited)
    recency_cutoff = target_recency
    rank_cutoff = max(
        num_strata_deposited - int(math.ceil(recency_cutoff)),
        0,
    )
    if num_strata_deposited == 0:
        assert rank_cutoff == 0
    else:
        assert 0 <= rank_cutoff <= num_strata_deposited - 1
    return rank_cutoff



from datetime import datetime
import typing

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_assign_contiguous_ids,
    alifestd_coerce_chronological_consistency,
    alifestd_find_chronological_inconsistency,
    alifestd_make_empty,
    alifestd_try_add_ancestor_id_col,
)

dt_epoch = datetime.utcfromtimestamp


@pytest.mark.parametrize("dtype", ["int", "str", "float", "object"])
def test_coerce_chronological_consistency_empty(dtype: str):
    df = alifestd_make_empty()
    df["origin_time"] = pd.Series(dtype=dtype)
    df_ = df.copy()
    assert alifestd_coerce_chronological_consistency(df).equals(df_)
    assert df.equals(df_)


def test_coerce_chronological_consistency_empty_datetime():
    df = alifestd_make_empty()
    df["origin_time"] = pd.to_datetime(pd.Series())
    df_ = df.copy()
    assert alifestd_coerce_chronological_consistency(df).equals(df_)
    assert df.equals(df_)


@pytest.mark.parametrize("wrap", [int, str, float, lambda x: (x,), dt_epoch])
def test_is_chronologically_ordered_sexual1(wrap: typing.Callable):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[333,44]",
                "[333,5]",
                "[333]",
                "[none]",
                "[333]",
                "[none]",
            ],
            "origin_time": [*map(wrap, [5, 5, 2, 1, 2, 0])],
        }
    )

    for phylogeny_df_ in phylogeny_df, alifestd_assign_contiguous_ids(
        phylogeny_df
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            assert alifestd_find_chronological_inconsistency(df) is None
            df_ = df.copy()
            corrected = alifestd_coerce_chronological_consistency(df)
            assert corrected.reset_index(drop=True).equals(
                df_.reset_index(drop=True),
            )
            assert df.equals(df_)


@pytest.mark.parametrize("wrap", [int, str, float, lambda x: (x,), dt_epoch])
def test_is_chronologically_ordered_sexual2(wrap: typing.Callable):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[333,44]",
                "[333,5]",
                "[333]",
                "[none]",
                "[333]",
                "[none]",
            ],
            "origin_time": [*map(wrap, [5, 5, 2, 1, 2, 2])],
        }
    )

    for phylogeny_df_ in phylogeny_df, alifestd_assign_contiguous_ids(
        phylogeny_df
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            assert alifestd_find_chronological_inconsistency(df) is None
            df_ = df.copy()
            corrected = alifestd_coerce_chronological_consistency(df)
            assert corrected.reset_index(drop=True).equals(
                df_.reset_index(drop=True),
            )
            assert df.equals(df_)


@pytest.mark.parametrize("wrap", [int, str, float, lambda x: (x,), dt_epoch])
def test_is_chronologically_unordered_sexual1(wrap: typing.Callable):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[333,44]",
                "[333,5]",
                "[333]",
                "[none]",
                "[333]",
                "[none]",
            ],
            "origin_time": [*map(wrap, [5, 5, 2, 1, 2, 6])],
        }
    )

    for phylogeny_df_ in phylogeny_df, alifestd_assign_contiguous_ids(
        phylogeny_df
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            assert alifestd_find_chronological_inconsistency(df) is not None
            df_ = df.copy()
            corrected = alifestd_coerce_chronological_consistency(df)
            assert not corrected.equals(df_)
            assert alifestd_find_chronological_inconsistency(corrected) is None
            assert df.equals(df_)


@pytest.mark.parametrize("wrap", [int, str, float, lambda x: (x,), dt_epoch])
def test_is_chronologically_unordered_sexual2(wrap: typing.Callable):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[333,44]",
                "[333,5]",
                "[333]",
                "[none]",
                "[333]",
                "[none]",
            ],
            "origin_time": [*map(wrap, [1, 5, 2, 1, 2, 2])],
        }
    )

    for phylogeny_df_ in (
        phylogeny_df,
        alifestd_assign_contiguous_ids(phylogeny_df),
        alifestd_try_add_ancestor_id_col(phylogeny_df),
        alifestd_try_add_ancestor_id_col(
            alifestd_assign_contiguous_ids(phylogeny_df)
        ),
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            assert alifestd_find_chronological_inconsistency(df) is not None
            df_ = df.copy()
            corrected = alifestd_coerce_chronological_consistency(df)
            assert not corrected.equals(df_)
            assert alifestd_find_chronological_inconsistency(corrected) is None
            assert df.equals(df_)


@pytest.mark.parametrize("wrap", [int, str, float, lambda x: (x,), dt_epoch])
def test_is_chronologically_unordered_sexual3(wrap: typing.Callable):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[333,44]",
                "[333,5]",
                "[333]",
                "[none]",
                "[333]",
                "[none]",
            ],
            "origin_time": [*map(wrap, [1, 5, 2, 3, 2, 2])],
        }
    )

    for phylogeny_df_ in (
        phylogeny_df,
        alifestd_assign_contiguous_ids(phylogeny_df),
        alifestd_try_add_ancestor_id_col(phylogeny_df),
        alifestd_try_add_ancestor_id_col(
            alifestd_assign_contiguous_ids(phylogeny_df)
        ),
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            assert alifestd_find_chronological_inconsistency(df) is not None
            df_ = df.copy()
            corrected = alifestd_coerce_chronological_consistency(df)
            assert not corrected.equals(df_)
            assert alifestd_find_chronological_inconsistency(corrected) is None
            assert df.equals(df_)


@pytest.mark.parametrize("wrap", [int, str, float, lambda x: (x,), dt_epoch])
def test_is_chronologically_ordered_asexual1(wrap: typing.Callable):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[44]",
                "[5]",
                "[333]",
                "[none]",
                "[1]",
                "[333]",
            ],
            "origin_time": [*map(wrap, [6, 5, 2, 1, 6, 2])],
        }
    )

    for phylogeny_df_ in (
        phylogeny_df,
        alifestd_assign_contiguous_ids(phylogeny_df),
        alifestd_try_add_ancestor_id_col(phylogeny_df),
        alifestd_try_add_ancestor_id_col(
            alifestd_assign_contiguous_ids(phylogeny_df)
        ),
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            assert alifestd_find_chronological_inconsistency(df) is None
            df_ = df.copy()
            corrected = alifestd_coerce_chronological_consistency(df)
            assert corrected.reset_index(drop=True).equals(
                df_.reset_index(drop=True),
            )
            assert df.equals(df_)


@pytest.mark.parametrize("wrap", [int, str, float, lambda x: (x,), dt_epoch])
def test_is_chronologically_ordered_asexual2(wrap: typing.Callable):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[44]",
                "[5]",
                "[333]",
                "[none]",
                "[1]",
                "[none]",
            ],
            "origin_time": [*map(wrap, [6, 5, 2, 1, 6, 2])],
        }
    )

    for phylogeny_df_ in (
        phylogeny_df,
        alifestd_assign_contiguous_ids(phylogeny_df),
        alifestd_try_add_ancestor_id_col(phylogeny_df),
        alifestd_try_add_ancestor_id_col(
            alifestd_assign_contiguous_ids(phylogeny_df)
        ),
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            assert alifestd_find_chronological_inconsistency(df) is None
            df_ = df.copy()
            corrected = alifestd_coerce_chronological_consistency(df)
            assert corrected.reset_index(drop=True).equals(
                df_.reset_index(drop=True),
            )
            assert df.equals(df_)


@pytest.mark.parametrize("wrap", [int, str, float, lambda x: (x,), dt_epoch])
def test_is_chronologically_unordered_asexual1(wrap: typing.Callable):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[333]",
                "[333]",
                "[333]",
                "[none]",
                "[333]",
                "[0]",
            ],
            "origin_time": [*map(wrap, [5, 5, 2, 1, 2, 4])],
        }
    )

    for phylogeny_df_ in (
        phylogeny_df,
        alifestd_assign_contiguous_ids(phylogeny_df),
        alifestd_try_add_ancestor_id_col(phylogeny_df),
        alifestd_try_add_ancestor_id_col(
            alifestd_assign_contiguous_ids(phylogeny_df)
        ),
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            assert alifestd_find_chronological_inconsistency(df) is not None
            df_ = df.copy()
            corrected = alifestd_coerce_chronological_consistency(df)
            assert not corrected.equals(df_)
            assert alifestd_find_chronological_inconsistency(corrected) is None
            assert df.equals(df_)


@pytest.mark.parametrize("wrap", [int, str, float, lambda x: (x,), dt_epoch])
def test_is_chronologically_unordered_asexual2(wrap: typing.Callable):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[44]",
                "[0]",
                "[333]",
                "[none]",
                "[333]",
                "[none]",
            ],
            "origin_time": [*map(wrap, [1, 5, 2, 1, 2, 2])],
        }
    )

    for phylogeny_df_ in (
        phylogeny_df,
        alifestd_assign_contiguous_ids(phylogeny_df),
        alifestd_try_add_ancestor_id_col(phylogeny_df),
        alifestd_try_add_ancestor_id_col(
            alifestd_assign_contiguous_ids(phylogeny_df)
        ),
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            assert alifestd_find_chronological_inconsistency(df) is not None
            df_ = df.copy()
            corrected = alifestd_coerce_chronological_consistency(df)
            assert not corrected.equals(df_)
            assert alifestd_find_chronological_inconsistency(corrected) is None
            assert df.equals(df_)


@pytest.mark.parametrize("wrap", [int, str, float, lambda x: (x,), dt_epoch])
def test_is_chronologically_unordered_asexual3(wrap: typing.Callable):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 22, 333, 44, 5],
            "ancestor_list": [
                "[333]",
                "[5]",
                "[44]",
                "[none]",
                "[333]",
                "[none]",
            ],
            "origin_time": [*map(wrap, [1, 5, 2, 3, 2, 2])],
        }
    )

    for phylogeny_df_ in (
        phylogeny_df,
        alifestd_assign_contiguous_ids(phylogeny_df),
        alifestd_try_add_ancestor_id_col(phylogeny_df),
        alifestd_try_add_ancestor_id_col(
            alifestd_assign_contiguous_ids(phylogeny_df)
        ),
    ):
        for df in phylogeny_df_, phylogeny_df_.sample(frac=1):
            assert alifestd_find_chronological_inconsistency(df) is not None
            df_ = df.copy()
            corrected = alifestd_coerce_chronological_consistency(df)
            assert not corrected.equals(df_)
            assert alifestd_find_chronological_inconsistency(corrected) is None
            assert df.equals(df_)



import pandas as pd


def alifestd_has_multiple_roots(phylogeny_df: pd.DataFrame) -> bool:
    """Does the phylogeny two or more root organisms?

    Input dataframe is not mutated by this operation.
    """
    return (
        (phylogeny_df["ancestor_list"] == "[]").sum()
        + (phylogeny_df["ancestor_list"] == "[NONE]").sum()
        + (phylogeny_df["ancestor_list"] == "[none]").sum()
        + (phylogeny_df["ancestor_list"] == "[None]").sum()
    ) >= 2



import typing  # pragma: no cover

from deprecated.sphinx import deprecated  # pragma: no cover
import numpy as np  # pragma: no cover
import pandas as pd  # pragma: no cover

from ..._auxiliary_lib import indices_of_unique  # pragma: no cover
from ._compile_phylogeny_from_lineage_iters import (  # pragma: no cover
    compile_phylogeny_from_lineage_iters,
)


# @MAM 02-04-2023
# created a partial implementation of missing features before testing showed
# it underperformed the GC tracker; left partial implementation here:
# https://gist.github.com/mmore500/06a359e528f59f3feb0c72dfc01b8fef
@deprecated(
    version="1.13.0", reason="Incompatible with numpy v2"
)  # pragma: no cover
class DecantingPhyloTracker:  # pragma: no cover
    """Data structure to enable perfect tracking over a fixed-size population
    with synchronous generations.

    Generally less performant than GarbageCollectingPhyloGracker; represents
    organism records as independent objects (which each require an independent
    allocations and, on lineage extinction, deletions). Uses a decanting buffer
    to reduce object creation overhead by only allocating record objects for
    organisms that produce extant lineages beyond a threshold length.

    Does not include organism population loc and trait values in phylogenetic
    record.
    """

    # static class constant
    _nan_val = -1

    @staticmethod
    def _is_nan(v: int) -> bool:
        return v < 0

    @staticmethod
    @np.vectorize
    def _tuple_wrap(v: typing.Any) -> typing.Tuple:
        return (v,)

    # decanting buffer layout
    #
    #                     population loc of extant organisms
    #                     |
    #                     |
    # generations ago ----|-> 0   1   2   3   4   ...   buffer_size - 1
    # from extant         V +----------------------------------------
    # organisms           0 |
    #                     1 |         POPULATION
    #                     2 |          POSITION
    #                     3 |             IDS
    #                     4 |
    #                   ... |
    #   population_size - 1 |
    #
    # layer 0 (z axis): organism's own population loc
    # layer 1 (z axis): parent's population loc
    #
    # every generation,
    # * the backtrack tree is updated using the subset population loc ids'
    #   that survived to the rightmost colun,
    # * population loc ids roll one column to the right,
    # * new population loc id's are pasted over column 0,
    # * and rows shuffle/duplicate according to the selected parent indices
    #
    _decanting_buffer: np.ndarray  # [int]
    _buffer_pos: int  # current start position of circular decanting buffer

    # permanent phylogeny storage after consolidation traversing decant buffer
    # using tuples instead of PerfectBacktrackHandle gives significant speedup
    _decanted_tree_tips: np.ndarray  # [typing.Tuple]

    def __init__(
        self: "DecantingPhyloTracker",
        population_size: int,
        buffer_size: int = 10,
        share_common_ancestor: bool = True,
    ) -> None:
        """Initialize data structure to perfectly track phylogenetic history of
        a population.

        Parameters
        ----------
        population_size : int
            How many locations are available within the tracked fixed-size
            population?
        buffer_size : int, optional
            How many generations should lineages be decanted before creating
            record objects for organisms with extant lineages?
        share_common_ancestor : bool, default True
            Should a dummy common ancestor be inserted as the first entry in
            the tracker?

            If True, all initial population members will be recorded as
            children of this dummy ancestor. If False, all initial
            population members will be recorded as having no parent.
        """

        if np.lib.NumpyVersion(np.__version__) >= "2.0.0b1":
            raise ImportError("This module is not compatible with numpy v2.")

        # initialize decanting buffer with all nan values
        self._decanting_buffer = np.full(
            (population_size, buffer_size, 2),
            self._nan_val,
            dtype=np.min_scalar_type(-population_size + 1),
        )
        self._buffer_pos = 0

        handles = []
        if share_common_ancestor:
            common_ancestor = (None,)
            # can't use list comprehension due to non-distinct tuple ids
            handles = []
            for __ in range(population_size):
                handle = (common_ancestor,)
                handles.append(handle)
            assert len(set(map(id, handles))) == len(handles)
            assert len(set(map(lambda x: id(x[0]), handles))) == 1
        else:
            # need to use mutable lists so ancestors have distinct id
            # because contents (i.e., None) are identical
            handles = [[None] for __ in range(population_size)]
            assert len(set(map(id, handles))) == len(handles)

        handles_array = np.empty(len(handles), dtype=object)
        handles_array[:] = handles
        self._decanted_tree_tips = handles_array
        assert len(set(map(id, self._decanted_tree_tips))) == len(handles)

    def _ArchiveBufferTail(
        self: "DecantingPhyloTracker",
    ):
        """Create record objects for decanted organisms with extant
        lineages."""
        assert not self._is_nan(self._decanting_buffer[0, self._buffer_pos, 0])

        # find locs of unique ancestors with extant descendants
        unique_row_indices = indices_of_unique(
            self._decanting_buffer[:, self._buffer_pos, 0],
        )

        unique_locs = self._decanting_buffer[
            unique_row_indices, self._buffer_pos, 0
        ]
        unique_parent_locs = self._decanting_buffer[
            unique_row_indices, self._buffer_pos, 1
        ]
        # apply unique ancestors' generational step
        # to relevant perfect tracking handles
        # step 1: copy parents
        self._decanted_tree_tips[unique_locs] = self._tuple_wrap(
            self._decanted_tree_tips[unique_parent_locs]
        )

    def _AdvanceBuffer(
        self: "DecantingPhyloTracker",
    ) -> None:
        """Progress circular decanting buffer, archiving buffer tail first if
        necessary."""
        # if rightmost column not empty, archive it before overwriting
        if not self._is_nan(self._decanting_buffer[0, self._buffer_pos, 0]):
            self._ArchiveBufferTail()

        # copy columns 0 thru buffer_size - 2 one column rightwards
        self._buffer_pos += 1
        # wrap buffer position around
        self._buffer_pos -= self._decanting_buffer.shape[1] * (
            self._buffer_pos >= self._decanting_buffer.shape[1]
        )

    def ElapseGeneration(
        self: "DecantingPhyloTracker",
        parent_locs: typing.List[int],
    ) -> None:
        """Append generational turnover to the phylogenetic record.

        Parameters
        ----------
        parent_locs : array_like of int
            Parent's population loc for each population loc.

            Position within array corresponds to post-turnover population
            members' population positions. Values within array correspond to
            those members' parents' population positions within the pre-
            turnover population.
        """
        # consolidate rows that now share common ancestry
        self._decanting_buffer = self._decanting_buffer[parent_locs, :, :]

        # shift buffer one position right,
        # archiving rightmost column if necessary
        self._AdvanceBuffer()

        # set new (leftmost) column
        # note: 0 - 1 == -1 is valid for indexing
        # layer 0: own population loc
        self._decanting_buffer[:, self._buffer_pos - 1, 0] = np.arange(
            self._decanting_buffer.shape[0]
        )
        # layer 1: parent population loc
        self._decanting_buffer[:, self._buffer_pos - 1, 1] = parent_locs

    def _FlushBuffer(self: "DecantingPhyloTracker") -> None:
        """Clear decanting buffer, creating record objects for all organisms
        currently progressing through the decanting process."""
        # advance buffer and fill column 0 with nan until empty
        # note: buffer may only be partway full, in which case
        # columns will shift rightwards until reaching rightmost column
        # and then begin to be archived
        while not np.all(self._decanting_buffer.flatten(), self._nan_val):
            self._AdvanceBuffer()
            self._decanting_buffer[:, self._buffer_pos - 1, :] = self._nan_val

    def CompilePhylogeny(
        self: "DecantingPhyloTracker",
        progress_wrap=lambda x: x,
    ) -> pd.DataFrame:
        """Generate full phylogenetic record of extant organisms.

        Parameters
        ----------
        progress_wrap : Callable, default identity function
            Wrapper applied around record row iterator; pass tqdm or equivalent
            to display progress bar for compilation process.

        Returns
        -------
        pandas.DataFrame
            Full phylogenetic record of extant organisms alife standard format.

        Notes
        -----
        This operation is non-destructive; further record-keeping may be
        performed on the tracker object afterwards.
        """
        self._FlushBuffer()
        assert set(self._decanting_buffer.flatten()) == {self._nan_val}

        def iter_lineage(
            tuple_handle: typing.Tuple,
        ) -> typing.Iterable[typing.Tuple]:
            while True:
                yield tuple_handle

                if tuple_handle[0] is None:
                    break
                else:
                    tuple_handle = tuple_handle[0]

        return compile_phylogeny_from_lineage_iters(
            iter_lineage(tip)
            for tip in progress_wrap(self._decanted_tree_tips)
        )



import typing

from ..._detail import PolicyCouplerBase
from ..._impl import CalcMrcaUncertaintyRelUpperBoundWorstCase
from .._PolicySpec import PolicySpec


class CalcMrcaUncertaintyRelUpperBound:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "CalcMrcaUncertaintyRelUpperBound",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(
        self: "CalcMrcaUncertaintyRelUpperBound",
        other: typing.Any,
    ) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcMrcaUncertaintyRelUpperBound",
        policy: PolicyCouplerBase,
        first_num_strata_deposited: int,
        second_num_strata_deposited: int,
        actual_rank_of_mrca: int,
    ) -> float:
        """At most, how much uncertainty to estimate rank of MRCA? Inclusive."""
        # rectify negative-indexed actual_rank_of_mrca
        if actual_rank_of_mrca is not None and actual_rank_of_mrca < 0:
            least_last_rank = min(
                first_num_strata_deposited - 1,
                second_num_strata_deposited - 1,
            )
            actual_rank_of_mrca += least_last_rank
            assert actual_rank_of_mrca >= 0

        spec = policy.GetSpec()

        if 0 in (first_num_strata_deposited, second_num_strata_deposited):
            return 0.0

        min_num_strata_deposited, max_num_strata_deposited = sorted(
            [
                first_num_strata_deposited,
                second_num_strata_deposited,
            ]
        )

        length_ratio = max_num_strata_deposited / min_num_strata_deposited

        if spec.GetRecencyProportionalResolution() == 0:
            return CalcMrcaUncertaintyRelUpperBoundWorstCase()(
                policy,
                first_num_strata_deposited,
                second_num_strata_deposited,
                actual_rank_of_mrca,
            )
        else:
            res = length_ratio / spec.GetRecencyProportionalResolution()

            # tighten to worst-possible case given number of strata deposited
            return min(
                res,
                CalcMrcaUncertaintyRelUpperBoundWorstCase()(
                    policy,
                    first_num_strata_deposited,
                    second_num_strata_deposited,
                    actual_rank_of_mrca,
                ),
            )



"""Implementation helpers."""

from ...._auxiliary_lib import lazy_attach_stub

__getattr__, __dir__, __all__ = lazy_attach_stub(
    __name__, __file__, should_launder=lambda _: False
)
del lazy_attach_stub



import typing

import numpy as np
import pandas as pd

from ._alifestd_categorize_triplet_asexual import (
    alifestd_categorize_triplet_asexual,
)
from ._alifestd_find_leaf_ids import alifestd_find_leaf_ids
from ._estimate_binomial_p import estimate_binomial_p


def alifestd_estimate_triplet_distance_asexual(
    first_df: pd.DataFrame,
    second_df: pd.DataFrame,
    taxon_label_key: str,
    confidence: float = 0.99,
    precision: float = 0.01,
    strict: typing.Union[bool, typing.Tuple[bool, bool]] = True,
    detail: bool = False,
    progress_wrap: typing.Callable = lambda x: x,
    mutate: bool = False,
) -> typing.Union[float, typing.Tuple[float, typing.Tuple[float, float, int]]]:
    """Estimate the triplet distance between two asexual phylogenetic trees in
    alife sampling sets of three leaf taxa and counting the fraction whose
    phylogenetic connectivity mismatch between trees.

    Parameters
    ----------
    first_df : pd.DataFrame
        The DataFrame representing the first phylogenetic tree.
    second_df : pd.DataFrame
        The DataFrame representing the second phylogenetic tree.
    taxon_label_key : str
        The key in the DataFrame to identify the taxon labels.
    confidence : float, default 0.99
        The confidence level for the estimation.

        See `estimate_binomial_p` for details.
    precision : float, default 0.01
        The precision of the estimation.

        See `estimate_binomial_p` for details.
    strict : bool or Tuple[bool, bool], default True
        A flag or a tuple of flags indicating how to treat tuples.

        If False, triplets that form a polytomy in either tree are not counted
        as mismatching. If True, they are counted as mismatching. If a tuple is
        given, polytomies in the first and second trees are treated according
        to the first and second elements of the tuple, respectively.
    detail : bool, default False
        If True, returns a detailed result including the estimated distance,
        confidence interval, and sample size.
    progress_wrap : typing.Callable, default lambda x: x
        A callable to display progress, e.g., tqdm.
    mutate : bool, default False
        If True, allows mutation of input DataFrames.

    Returns
    -------
    float or Tuple[float, Tuple[float, float, int]]
        The estimated distance between the two trees.

        If `detail` is True, returns a tuple containing the estimated distance,
        the confidence interval, and the sample size.

    Notes
    -----
    The core comparison is done by sampling triplets of taxa, categorizing
    them, and comparing these categorizations across the two trees, taking into
    account the `strict` and `lax` parameters for handling polytomies. See
    `alifestd_categorize_triplet_asexual` for details.

    See Also
    --------
    alifestd_categorize_triplet_asexual
    alifestd_sample_triplet_comparisons_asexual
    """
    if not mutate:
        first_df = first_df.copy()
        second_df = second_df.copy()
    try:
        strict1, strict2 = strict
    except TypeError:
        strict1 = strict2 = strict
        assert strict1 == strict2 == strict
    lax = (not bool(strict1), not bool(strict2))

    first_df = first_df.reset_index(drop=True).set_index("id", drop=False)
    second_df = second_df.reset_index(drop=True).set_index("id", drop=False)

    first_leaf_ids = alifestd_find_leaf_ids(first_df)
    first_taxon_labels = first_df.loc[first_leaf_ids, taxon_label_key]
    second_leaf_ids = alifestd_find_leaf_ids(second_df)
    second_taxon_labels = second_df.loc[second_leaf_ids, taxon_label_key]

    assert first_taxon_labels.is_unique
    assert second_taxon_labels.is_unique
    assert (
        first_taxon_labels.sort_values().to_numpy()
        == second_taxon_labels.sort_values().to_numpy()
    ).all()

    if len(first_leaf_ids) < 3:
        return (0.0, [0.0, 0.0], 0) if detail else 0.0

    def sample_triplet_comparison() -> bool:
        """Return True if the two categorizations differ."""
        labels = np.random.choice(first_taxon_labels, 3, replace=False)
        first_triplet_ids = [
            first_df.index[first_df[taxon_label_key] == label].item()
            for label in labels
        ]
        second_triplet_ids = [
            second_df.index[second_df[taxon_label_key] == label].item()
            for label in labels
        ]
        assert 3 == len(first_triplet_ids) == len(second_triplet_ids)

        cat1 = alifestd_categorize_triplet_asexual(
            first_df, first_triplet_ids, mutate=True
        )
        cat2 = alifestd_categorize_triplet_asexual(
            second_df, second_triplet_ids, mutate=True
        )

        polytomy = -1
        if cat1 == cat2:
            return False
        elif lax[0] and cat1 == polytomy:
            return False
        elif lax[1] and cat2 == polytomy:
            return False
        else:
            return True

    res = estimate_binomial_p(
        sample_triplet_comparison,
        confidence=confidence,
        precision=precision,
        progress_wrap=progress_wrap,
    )
    return res if detail else res[0]



import itertools as it

import pytest

from hstrat import hstrat


def test_CalcDefinitiveMinRanksSinceFirstRetainedDisparityWith_specimen():
    column = hstrat.HereditaryStratigraphicColumn()
    column2 = hstrat.HereditaryStratigraphicColumn()
    column.DepositStrata(100)

    child1 = column.CloneDescendant()
    child2 = column.CloneDescendant()

    assert hstrat.calc_definitive_min_ranks_since_first_retained_disparity_with(
        hstrat.col_to_specimen(column), hstrat.col_to_specimen(column2)
    ) == hstrat.calc_definitive_min_ranks_since_first_retained_disparity_with(
        column, column2
    )

    assert hstrat.calc_definitive_min_ranks_since_first_retained_disparity_with(
        hstrat.col_to_specimen(column), hstrat.col_to_specimen(column)
    ) == hstrat.calc_definitive_min_ranks_since_first_retained_disparity_with(
        column, column
    )

    assert hstrat.calc_definitive_min_ranks_since_first_retained_disparity_with(
        hstrat.col_to_specimen(column), hstrat.col_to_specimen(child1)
    ) == hstrat.calc_definitive_min_ranks_since_first_retained_disparity_with(
        column, child1
    )

    assert hstrat.calc_definitive_min_ranks_since_first_retained_disparity_with(
        hstrat.col_to_specimen(child1), hstrat.col_to_specimen(child2)
    ) == hstrat.calc_definitive_min_ranks_since_first_retained_disparity_with(
        child1, child2
    )

    child1.DepositStrata(10)
    assert hstrat.calc_definitive_min_ranks_since_first_retained_disparity_with(
        hstrat.col_to_specimen(child1), hstrat.col_to_specimen(child2)
    ) == hstrat.calc_definitive_min_ranks_since_first_retained_disparity_with(
        child1, child2
    )


@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8],
)
def test_CalcDefinitiveMinRanksSinceFirstRetainedDisparityWith1(
    ordered_store,
    differentia_width,
):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=differentia_width,
        stratum_ordered_store=ordered_store,
    )

    column.DepositStrata(100)

    offspring1 = column.CloneDescendant()
    offspring2 = column.CloneDescendant()

    for c1, c2 in it.combinations([column, offspring1, offspring2], 2):
        if differentia_width == 64:
            assert (
                hstrat.calc_definitive_min_ranks_since_first_retained_disparity_with(
                    c1,
                    c2,
                )
                == c1.GetNumStrataDeposited()
                - column.GetNumStrataDeposited()
                - 1
            )
            assert (
                hstrat.calc_definitive_min_ranks_since_first_retained_disparity_with(
                    c2,
                    c1,
                )
                == c2.GetNumStrataDeposited()
                - column.GetNumStrataDeposited()
                - 1
            )
        else:
            assert hstrat.calc_definitive_min_ranks_since_first_retained_disparity_with(
                c1,
                c2,
            ) in (
                c1.GetNumStrataDeposited()
                - column.GetNumStrataDeposited()
                - 1,
                None,
            )
            assert hstrat.calc_definitive_min_ranks_since_first_retained_disparity_with(
                c2,
                c1,
            ) in (
                c2.GetNumStrataDeposited()
                - column.GetNumStrataDeposited()
                - 1,
                None,
            )

    for c in [column, offspring1, offspring2]:
        assert (
            hstrat.calc_definitive_min_ranks_since_first_retained_disparity_with(
                c, c
            )
            is None
        )

    offspring1.DepositStrata(100)
    offspring2.DepositStrata(100)

    for c1, c2 in it.combinations([column, offspring1, offspring2], 2):
        if differentia_width == 64:
            assert (
                hstrat.calc_definitive_min_ranks_since_first_retained_disparity_with(
                    c1,
                    c2,
                )
                == c1.GetNumStrataDeposited()
                - column.GetNumStrataDeposited()
                - 1
            )
            assert (
                hstrat.calc_definitive_min_ranks_since_first_retained_disparity_with(
                    c2,
                    c1,
                )
                == c2.GetNumStrataDeposited()
                - column.GetNumStrataDeposited()
                - 1
            )
        else:
            res = hstrat.calc_definitive_min_ranks_since_first_retained_disparity_with(
                c1,
                c2,
            )
            assert (
                res
                <= c1.GetNumStrataDeposited()
                - column.GetNumStrataDeposited()
                - 1
            )

    for c in [column, offspring1, offspring2]:
        assert (
            hstrat.calc_definitive_min_ranks_since_first_retained_disparity_with(
                c, c
            )
            is None
        )


@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "policy",
    [
        hstrat.fixed_resolution_algo.Policy(3),
        hstrat.recency_proportional_resolution_algo.Policy(1),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
def test_artifact_types_equiv(differentia_width, policy):
    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c1 = common_ancestor.CloneNthDescendant(4)
    c2 = common_ancestor.CloneNthDescendant(9)
    c_x = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c_y = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    )

    for a, b in it.product(
        [common_ancestor, c1, c2, c_x, c_y],
        [common_ancestor, c1, c2, c_x, c_y],
    ):
        assert hstrat.calc_definitive_min_ranks_since_first_retained_disparity_with(
            hstrat.col_to_specimen(a),
            hstrat.col_to_specimen(b),
        ) == hstrat.calc_definitive_min_ranks_since_first_retained_disparity_with(
            a, b
        )



from ...._auxiliary_lib import jit


@jit(nopython=True)
def _get_island_id(
    population_loc: int,
    island_size: int,
) -> int:
    """Get population slot's island ID."""
    # islands are contiguous and sequential within population
    return population_loc // island_size



from hstrat._auxiliary_lib import textwrap_respect_indents


def test_simple_wrapping():
    text = "This is a simple test to wrap text appropriately."
    result = textwrap_respect_indents(text, ncol=20)
    expected = """This is a simple
test to wrap text
appropriately."""
    assert result == expected


def test_preserve_indentation():
    text = "    Indented line that should respect indentation when wrapped."
    result = textwrap_respect_indents(text, ncol=20)
    expected = """    Indented line
    that should respect
    indentation when
    wrapped."""
    assert result == expected


def test_empty_line_handling():
    text = "First line\n\nThird line"
    result = textwrap_respect_indents(text, ncol=20)
    expected = """First line

Third line"""
    assert result == expected


def test_long_unindented_line():
    text = "ThisIsAVeryLongSingleWordWithoutSpacesThatShouldNotWrap"
    result = textwrap_respect_indents(text, ncol=1000)
    expected = text
    assert result == expected


def test_custom_width():
    text = "Custom width wrapping test."
    result = textwrap_respect_indents(text, ncol=10)
    expected = """Custom
width
wrapping
test."""
    assert result == expected



import itertools as it

import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64],
)
def test_calc_rank_mrca_bounds_provided_confidence_level_specimen(
    retention_policy, differentia_width
):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column2 = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column.DepositStrata(100)

    child1 = column.CloneDescendant()
    child2 = column.CloneDescendant()

    assert hstrat.calc_rank_of_mrca_bounds_provided_confidence_level(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(column2),
        prior="arbitrary",
    ) == hstrat.calc_rank_of_mrca_bounds_provided_confidence_level(
        column, column2, prior="arbitrary"
    )

    assert hstrat.calc_rank_of_mrca_bounds_provided_confidence_level(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(column),
        prior="arbitrary",
    ) == hstrat.calc_rank_of_mrca_bounds_provided_confidence_level(
        column, column, prior="arbitrary"
    )

    assert hstrat.calc_rank_of_mrca_bounds_provided_confidence_level(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(child1),
        prior="arbitrary",
    ) == hstrat.calc_rank_of_mrca_bounds_provided_confidence_level(
        column, child1, prior="arbitrary"
    )

    assert hstrat.calc_rank_of_mrca_bounds_provided_confidence_level(
        hstrat.col_to_specimen(child1),
        hstrat.col_to_specimen(child2),
        prior="arbitrary",
    ) == hstrat.calc_rank_of_mrca_bounds_provided_confidence_level(
        child1, child2, prior="arbitrary"
    )

    child1.DepositStrata(10)
    assert hstrat.calc_rank_of_mrca_bounds_provided_confidence_level(
        hstrat.col_to_specimen(child1),
        hstrat.col_to_specimen(child2),
        prior="arbitrary",
    ) == hstrat.calc_rank_of_mrca_bounds_provided_confidence_level(
        child1, child2, prior="arbitrary"
    )


def test_CalcRankOfMrcaBoundsProvidedConfidenceLevel():
    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
    )
    assert (
        hstrat.calc_rank_of_mrca_bounds_provided_confidence_level(
            c1, c1, "arbitrary", 0.5
        )
        == 0.5
    )
    assert (
        hstrat.calc_rank_of_mrca_bounds_provided_confidence_level(
            c1, c1, "arbitrary", 0.6
        )
        == 0.75
    )
    assert (
        hstrat.calc_rank_of_mrca_bounds_provided_confidence_level(
            c1, c1, "arbitrary", 0.75
        )
        == 0.75
    )

    c2 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
    )
    p = 1 / 2**64
    assert (
        hstrat.calc_rank_of_mrca_bounds_provided_confidence_level(
            c2, c2, "arbitrary", 0.5
        )
        == 1 - p
    )
    assert (
        hstrat.calc_rank_of_mrca_bounds_provided_confidence_level(
            c2, c2, "arbitrary", 0.6
        )
        == 1 - p
    )
    assert (
        hstrat.calc_rank_of_mrca_bounds_provided_confidence_level(
            c2, c2, "arbitrary", 0.75
        )
        == 1 - p
    )
    assert (
        hstrat.calc_rank_of_mrca_bounds_provided_confidence_level(
            c2, c2, "arbitrary", 0.95
        )
        == 1 - p
    )
    assert (
        hstrat.calc_rank_of_mrca_bounds_provided_confidence_level(
            c2, c2, "arbitrary", 0.99
        )
        == 1 - p
    )


@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "policy",
    [
        hstrat.fixed_resolution_algo.Policy(3),
        hstrat.recency_proportional_resolution_algo.Policy(1),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
@pytest.mark.parametrize(
    "prior",
    ["arbitrary"],
)
def test_artifact_types_equiv(differentia_width, policy, prior):
    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c1 = common_ancestor.CloneNthDescendant(4)
    c2 = common_ancestor.CloneNthDescendant(9)
    c_x = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c_y = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    )

    for a, b in it.product(
        [common_ancestor, c1, c2, c_x, c_y],
        [common_ancestor, c1, c2, c_x, c_y],
    ):
        assert hstrat.calc_rank_of_mrca_bounds_provided_confidence_level(
            hstrat.col_to_specimen(a),
            hstrat.col_to_specimen(b),
            prior,
        ) == hstrat.calc_rank_of_mrca_bounds_provided_confidence_level(
            a, b, prior
        )



import itertools as it
import random

import numpy as np
import pytest
from scipy import stats

from hstrat import hstrat


@pytest.mark.parametrize(
    "estimator",
    [
        "maximum_likelihood",
        "unbiased",
    ],
)
@pytest.mark.parametrize(
    "prior",
    [
        "arbitrary",
        hstrat.GeometricPrior(1.1),
        hstrat.ExponentialPrior(1.1),
        "uniform",
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64],
)
def test_estimate_patristic_distance_between_specimen(
    estimator, prior, retention_policy, differentia_width
):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column2 = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column.DepositStrata(100)

    child1 = column.CloneDescendant()
    child2 = column.CloneDescendant()

    assert hstrat.estimate_patristic_distance_between(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(column2),
        estimator=estimator,
        prior=prior,
    ) == hstrat.estimate_patristic_distance_between(
        column,
        column2,
        estimator=estimator,
        prior=prior,
    )

    assert hstrat.estimate_patristic_distance_between(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(column),
        estimator=estimator,
        prior=prior,
    ) == hstrat.estimate_patristic_distance_between(
        column,
        column,
        estimator=estimator,
        prior=prior,
    )

    assert hstrat.estimate_patristic_distance_between(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(child1),
        estimator=estimator,
        prior=prior,
    ) == hstrat.estimate_patristic_distance_between(
        column,
        child1,
        estimator=estimator,
        prior=prior,
    )

    assert hstrat.estimate_patristic_distance_between(
        hstrat.col_to_specimen(child1),
        hstrat.col_to_specimen(child2),
        estimator=estimator,
        prior=prior,
    ) == hstrat.estimate_patristic_distance_between(
        child1,
        child2,
        estimator=estimator,
        prior=prior,
    )

    child1.DepositStrata(10)
    assert hstrat.estimate_patristic_distance_between(
        hstrat.col_to_specimen(child1),
        hstrat.col_to_specimen(child2),
        estimator=estimator,
        prior=prior,
    ) == hstrat.estimate_patristic_distance_between(
        child1,
        child2,
        estimator=estimator,
        prior=prior,
    )


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "estimator",
    [
        "maximum_likelihood",
        "unbiased",
    ],
)
@pytest.mark.parametrize(
    "prior",
    [
        "arbitrary",
        hstrat.GeometricPrior(1.1),
        hstrat.ExponentialPrior(1.1),
        "uniform",
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
def test_comparison_commutativity_asynchronous(
    differentia_width,
    retention_policy,
    estimator,
    prior,
):
    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(30):
        _ = _generation
        for first, second in it.combinations(population, 2):
            # assert commutativity
            assert hstrat.estimate_patristic_distance_between(
                first,
                second,
                estimator=estimator,
                prior=prior,
            ) == hstrat.estimate_patristic_distance_between(
                second,
                first,
                estimator=estimator,
                prior=prior,
            )

        # advance generation
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].CloneDescendant()
        for individual in population:
            # asynchronous generations
            if random.choice([True, False]):
                individual.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        pytest.param(
            hstrat.perfect_resolution_algo.Policy(),
            marks=pytest.mark.heavy_2a,
        ),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "estimator",
    ["maximum_likelihood", "unbiased"],
)
@pytest.mark.parametrize(
    "prior",
    [
        "arbitrary",
        hstrat.GeometricPrior(1.1),
        hstrat.ExponentialPrior(1.1),
        "uniform",
    ],
)
def test_comparison_commutativity_synchronous(
    retention_policy,
    differentia_width,
    estimator,
    prior,
):

    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(30):
        _ = _generation
        for first, second in it.combinations(population, 2):
            # assert commutativity
            assert hstrat.estimate_patristic_distance_between(
                first,
                second,
                estimator=estimator,
                prior=prior,
            ) == hstrat.estimate_patristic_distance_between(
                second,
                first,
                estimator=estimator,
                prior=prior,
            )

        # advance generation
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].Clone()
        # synchronous generations
        for individual in population:
            individual.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        pytest.param(
            hstrat.perfect_resolution_algo.Policy(),
            marks=pytest.mark.heavy,
        ),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "estimator",
    ["maximum_likelihood", "unbiased"],
)
@pytest.mark.parametrize(
    "prior",
    [
        "arbitrary",
        hstrat.GeometricPrior(1.1),
        hstrat.ExponentialPrior(1.1),
        "uniform",
    ],
)
def test_comparison_validity(
    retention_policy,
    differentia_width,
    estimator,
    prior,
):
    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(30):
        _ = _generation
        for first, second in it.permutations(population, 2):

            rsmw1 = hstrat.estimate_ranks_since_mrca_with(
                first,
                second,
                estimator=estimator,
                prior=prior,
            )
            rsmw2 = hstrat.estimate_ranks_since_mrca_with(
                second,
                first,
                estimator=estimator,
                prior=prior,
            )
            est = hstrat.estimate_patristic_distance_between(
                first,
                second,
                estimator=estimator,
                prior=prior,
            )
            assert (rsmw1 is None) == (rsmw1 is None) == (est is None)
            if est is not None:
                assert rsmw1 + rsmw2 == pytest.approx(est)

        # advance generations asynchronously
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].Clone()
        for individual in population:
            if random.choice([True, False]):
                individual.DepositStratum()


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=13),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
def test_statistical_properties(
    retention_policy,
    differentia_width,
):

    err_maximum_likelihood_arbitrary = []
    err_maximum_likelihood_uniform = []
    err_unbiased = []

    common_ancestors = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
        )
    ]
    for __ in range(113):
        common_ancestors.append(common_ancestors[-1].CloneDescendant())

    for _rep in range(10000):
        _ = _rep
        num_total = random.randrange(57, 113)
        num_together = random.randrange(num_total + 1)
        num_alone = num_total - num_together

        left_alone = num_alone
        right_alone = num_alone + random.randrange(19)

        common_ancestor = common_ancestors[num_together]
        left = common_ancestor.Clone()
        right = common_ancestor.Clone()

        left.DepositStrata(left_alone)
        right.DepositStrata(right_alone)

        err_maximum_likelihood_arbitrary.append(
            hstrat.estimate_patristic_distance_between(
                left, right, estimator="maximum_likelihood", prior="arbitrary"
            )
            - (left_alone + right_alone)
        )
        err_maximum_likelihood_uniform.append(
            hstrat.estimate_patristic_distance_between(
                left, right, estimator="maximum_likelihood", prior="uniform"
            )
            - (left_alone + right_alone)
        )
        err_unbiased.append(
            hstrat.estimate_patristic_distance_between(
                left, right, estimator="unbiased", prior="uniform"
            )
            - (left_alone + right_alone)
        )

    err_unbiased_ = np.array(err_unbiased)
    print(np.mean(err_unbiased))
    assert stats.ttest_ind(err_unbiased_, -err_unbiased_)[1] > 0.01

    for err_maximum_likelihood in (
        err_maximum_likelihood_arbitrary,
        err_maximum_likelihood_uniform,
    ):
        mean_err_unbiased = np.mean(err_unbiased)
        mean_err_maximum_likelihood = np.mean(err_maximum_likelihood)

        median_abs_err_unbiased = np.percentile(np.abs(err_unbiased), 50)
        median_abs_err_maximum_likelihood = np.percentile(
            np.abs(err_maximum_likelihood), 50
        )

        if differentia_width == 1 and isinstance(
            retention_policy, hstrat.nominal_resolution_algo.Policy
        ):
            assert abs(mean_err_unbiased) < abs(mean_err_maximum_likelihood)
            assert not all(
                ml > ub
                for ml, ub in zip(
                    sorted(np.abs(err_maximum_likelihood)),
                    sorted(np.abs(err_unbiased)),
                )
            )
        elif differentia_width == 1:
            assert abs(mean_err_unbiased) < abs(mean_err_maximum_likelihood)
            assert median_abs_err_unbiased > median_abs_err_maximum_likelihood
        else:
            assert abs(mean_err_unbiased) <= abs(mean_err_maximum_likelihood)
            assert median_abs_err_unbiased >= median_abs_err_maximum_likelihood


@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "policy",
    [
        hstrat.fixed_resolution_algo.Policy(3),
        hstrat.recency_proportional_resolution_algo.Policy(1),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
@pytest.mark.parametrize(
    "estimator",
    ["maximum_likelihood", "unbiased"],
)
@pytest.mark.parametrize(
    "prior",
    ["arbitrary", "uniform"],
)
def test_artifact_types_equiv(differentia_width, policy, estimator, prior):
    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c1 = common_ancestor.CloneNthDescendant(4)
    c2 = common_ancestor.CloneNthDescendant(9)
    c_x = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c_y = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    )

    for a, b in it.product(
        [common_ancestor, c1, c2, c_x, c_y],
        [common_ancestor, c1, c2, c_x, c_y],
    ):
        assert hstrat.estimate_patristic_distance_between(
            hstrat.col_to_specimen(a),
            hstrat.col_to_specimen(b),
            estimator,
            prior,
        ) == hstrat.estimate_patristic_distance_between(a, b, estimator, prior)



import typing
import warnings

from ._is_in_coverage_run import is_in_coverage_run


class _ShimFtor:
    __wrapped__: typing.Callable  # shim for numba jit attr

    def __init__(self: "_ShimFtor", wrapped: typing.Callable) -> None:
        self.__wrapped__ = wrapped

    def __call__(self: "_ShimFtor", *args, **kwargs) -> typing.Any:
        return self.__wrapped__(*args, **kwargs)


def jit(*args, **kwargs) -> typing.Callable:
    """Decorator that performs jit compilation if numba available.

    Disables jit during coverage measurement to increase source
    visibility.
    """
    try:
        import numba as nb
    except (ImportError, ModuleNotFoundError):  # pragma: no cover
        warnings.warn(
            "numba unavailable,"
            "wrapped function may lose significant performance. "
            "To get numba, install it directly or install hstrat optional jit "
            "extras: python -m pip install hstrat[jit].",
            ImportWarning,
        )
        return _ShimFtor

    if is_in_coverage_run():
        warnings.warn(
            "code coverage tracing detected,"
            "disabling jit compilation to increase source visibility",
            RuntimeWarning,
        )
        return _ShimFtor
    else:  # pragma: no cover
        # exclude from coverage because jit compilation disabled in cov runs
        return nb.jit(*args, **kwargs)



import itertools as it
import random

import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "confidence_level",
    [0.95, 0.88],
)
def test_calc_rank_of_mrca_uncertainty_between_specimen(
    retention_policy, differentia_width, confidence_level
):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column2 = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column.DepositStrata(100)

    child1 = column.CloneDescendant()
    child2 = column.CloneDescendant()

    assert hstrat.calc_rank_of_mrca_uncertainty_between(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(column2),
        prior="arbitrary",
        confidence_level=confidence_level,
    ) == hstrat.calc_rank_of_mrca_uncertainty_between(
        column, column2, prior="arbitrary", confidence_level=confidence_level
    )

    assert hstrat.calc_rank_of_mrca_uncertainty_between(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(column),
        prior="arbitrary",
        confidence_level=confidence_level,
    ) == hstrat.calc_rank_of_mrca_uncertainty_between(
        column, column, prior="arbitrary", confidence_level=confidence_level
    )

    assert hstrat.calc_rank_of_mrca_uncertainty_between(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(child1),
        prior="arbitrary",
        confidence_level=confidence_level,
    ) == hstrat.calc_rank_of_mrca_uncertainty_between(
        column, child1, prior="arbitrary", confidence_level=confidence_level
    )

    assert hstrat.calc_rank_of_mrca_uncertainty_between(
        hstrat.col_to_specimen(child1),
        hstrat.col_to_specimen(child2),
        prior="arbitrary",
        confidence_level=confidence_level,
    ) == hstrat.calc_rank_of_mrca_uncertainty_between(
        child1, child2, prior="arbitrary", confidence_level=confidence_level
    )

    child1.DepositStrata(10)
    assert hstrat.calc_rank_of_mrca_uncertainty_between(
        hstrat.col_to_specimen(child1),
        hstrat.col_to_specimen(child2),
        prior="arbitrary",
        confidence_level=confidence_level,
    ) == hstrat.calc_rank_of_mrca_uncertainty_between(
        child1, child2, prior="arbitrary", confidence_level=confidence_level
    )


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreTree,
            marks=pytest.mark.heavy,
        ),
    ],
)
def test_comparison_commutativity_asynchronous(
    retention_policy,
    ordered_store,
):
    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store=ordered_store,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(100):
        _ = _generation
        for first, second in it.combinations(population, 2):
            # assert commutativity
            assert hstrat.calc_rank_of_mrca_uncertainty_between(
                first, second, prior="arbitrary"
            ) == hstrat.calc_rank_of_mrca_uncertainty_between(
                second, first, prior="arbitrary"
            )

        # advance generation
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].CloneDescendant()
        for individual in population:
            # asynchronous generations
            if random.choice([True, False]):
                individual.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        pytest.param(
            hstrat.perfect_resolution_algo.Policy(),
            marks=pytest.mark.heavy_2a,
        ),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreDict,
            marks=pytest.mark.heavy_2b,
        ),
        hstrat.HereditaryStratumOrderedStoreList,
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreTree,
            marks=pytest.mark.heavy,
        ),
    ],
)
def test_comparison_commutativity_synchronous(
    retention_policy,
    ordered_store,
):

    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store=ordered_store,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(100):
        _ = _generation
        for first, second in it.combinations(population, 2):
            # assert commutativity
            assert hstrat.calc_rank_of_mrca_uncertainty_between(
                first, second, prior="arbitrary"
            ) == hstrat.calc_rank_of_mrca_uncertainty_between(
                second, first, prior="arbitrary"
            )

        # advance generation
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].Clone()
        # synchronous generations
        for individual in population:
            individual.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreTree,
            marks=pytest.mark.heavy,
        ),
    ],
)
def test_comparison_validity(retention_policy, ordered_store):
    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store=ordered_store,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(100):
        _ = _generation
        for first, second in it.combinations(population, 2):
            assert (
                hstrat.calc_rank_of_mrca_uncertainty_between(
                    first, second, prior="arbitrary"
                )
                >= 0
            )

        # advance generations asynchronously
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].Clone()
        for individual in population:
            if random.choice([True, False]):
                individual.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy1",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "retention_policy2",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_scenario_no_mrca(
    retention_policy1,
    retention_policy2,
    ordered_store,
):
    first = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy1,
    )
    second = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy2,
    )

    for _generation in range(100):
        _ = _generation
        assert (
            hstrat.calc_rank_of_mrca_uncertainty_between(
                first, second, prior="arbitrary"
            )
            == 0
        )
        assert (
            hstrat.calc_rank_of_mrca_uncertainty_between(
                second, first, prior="arbitrary"
            )
            == 0
        )

        first.DepositStratum()
        second.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_scenario_no_divergence(retention_policy, ordered_store):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )

    for _generation in range(100):
        _ = _generation
        assert (
            hstrat.calc_rank_of_mrca_uncertainty_between(
                column, column, prior="arbitrary"
            )
            == 0
        )

        column.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "confidence_level",
    [0.8, 0.95],
)
def test_CalcRankOfMrcaUncertaintyWith_narrow_shallow(
    retention_policy,
    differentia_width,
    confidence_level,
):

    columns = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
            stratum_ordered_store=hstrat.HereditaryStratumOrderedStoreDict,
        )
        for __ in range(20)
    ]

    steps = list(
        range(
            columns[
                0
            ].CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                significance_level=1 - confidence_level,
            )
            - columns[0].GetNumStrataDeposited()
        )
    )

    for step1, step2 in it.product(steps, steps):
        column1 = [col.Clone() for col in columns]
        column2 = [col.Clone() for col in columns]
        for __ in range(step1):
            for col in column1:
                col.DepositStratum()
        for i in range(step2):
            for col in column2:
                col.DepositStratum()

        for c1, c2 in zip(column1, column2):
            assert (
                hstrat.calc_rank_of_mrca_uncertainty_between(
                    c1,
                    c2,
                    prior="arbitrary",
                    confidence_level=confidence_level,
                )
                is None
            )
            assert (
                hstrat.calc_rank_of_mrca_uncertainty_between(
                    c2,
                    c1,
                    prior="arbitrary",
                    confidence_level=confidence_level,
                )
                is None
            )


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "confidence_level",
    [0.8, 0.95],
)
@pytest.mark.parametrize(
    "mrca_rank",
    [100],
)
def test_CalcRankOfMrcaUncertaintyWith_narrow_with_mrca(
    retention_policy,
    differentia_width,
    confidence_level,
    mrca_rank,
):

    columns = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
            stratum_ordered_store=hstrat.HereditaryStratumOrderedStoreDict,
        )
        for __ in range(20)
    ]

    for column in columns:
        column.DepositStrata(mrca_rank)

    steps = (0, 16, 51)

    for step1, step2 in it.product(steps, steps):
        column1 = [col.Clone() for col in columns]
        column2 = [col.Clone() for col in columns]
        for __ in range(step1):
            for col in column1:
                col.DepositStratum()
        for i in range(step2):
            for col in column2:
                col.DepositStratum()

        for c1, c2 in zip(column1, column2):
            assert hstrat.calc_rank_of_mrca_uncertainty_between(
                c1, c2, prior="arbitrary", confidence_level=confidence_level
            ) == hstrat.calc_rank_of_mrca_uncertainty_between(
                c2,
                c1,
                prior="arbitrary",
                confidence_level=confidence_level,
            )
            res = hstrat.calc_rank_of_mrca_uncertainty_between(
                c1,
                c2,
                prior="arbitrary",
                confidence_level=confidence_level,
            )

            if res is not None:
                assert res >= 0
                assert hstrat.calc_rank_of_mrca_uncertainty_between(
                    c1,
                    c2,
                    prior="arbitrary",
                    confidence_level=confidence_level,
                ) >= hstrat.calc_rank_of_mrca_uncertainty_between(
                    c2,
                    c1,
                    prior="arbitrary",
                    confidence_level=confidence_level / 2,
                )


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "confidence_level",
    [0.8, 0.95],
)
@pytest.mark.parametrize(
    "mrca_rank",
    [0, 100],
)
def test_CalcRankOfMrcaUncertaintyWith_narrow_no_mrca(
    retention_policy,
    differentia_width,
    confidence_level,
    mrca_rank,
):
    def make_column():
        return hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
            stratum_ordered_store=hstrat.HereditaryStratumOrderedStoreDict,
        )

    columns = [make_column() for __ in range(20)]

    for column in columns:
        column.DepositStrata(mrca_rank)

    steps = (0, 16, 51)

    for step1, step2 in it.product(steps, steps):
        column1 = [make_column() for col in columns]
        column2 = [make_column() for col in columns]
        for __ in range(step1):
            for col in column1:
                col.DepositStratum()
        for i in range(step2):
            for col in column2:
                col.DepositStratum()

        for c1, c2 in zip(column1, column2):
            assert hstrat.calc_rank_of_mrca_uncertainty_between(
                c1, c2, prior="arbitrary", confidence_level=confidence_level
            ) == hstrat.calc_rank_of_mrca_uncertainty_between(
                c2,
                c1,
                prior="arbitrary",
                confidence_level=confidence_level,
            )
            res = hstrat.calc_rank_of_mrca_uncertainty_between(
                c1,
                c2,
                prior="arbitrary",
                confidence_level=confidence_level,
            )

            if res is not None:
                assert 0 <= res
                assert (
                    hstrat.calc_rank_of_mrca_uncertainty_between(
                        c1,
                        c2,
                        prior="arbitrary",
                        confidence_level=confidence_level,
                    )
                    >= hstrat.calc_rank_of_mrca_uncertainty_between(
                        c2,
                        c1,
                        prior="arbitrary",
                        confidence_level=confidence_level / 2,
                    )
                    or hstrat.calc_rank_of_mrca_uncertainty_between(
                        c1,
                        c2,
                        prior="arbitrary",
                        confidence_level=confidence_level,
                    )
                    == 0
                )


@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "policy",
    [
        hstrat.fixed_resolution_algo.Policy(3),
        hstrat.recency_proportional_resolution_algo.Policy(1),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
@pytest.mark.parametrize(
    "prior",
    ["arbitrary"],
)
def test_artifact_types_equiv(differentia_width, policy, prior):
    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c1 = common_ancestor.CloneNthDescendant(4)
    c2 = common_ancestor.CloneNthDescendant(9)
    c_x = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c_y = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    )

    for a, b in it.product(
        [common_ancestor, c1, c2, c_x, c_y],
        [common_ancestor, c1, c2, c_x, c_y],
    ):
        assert hstrat.calc_rank_of_mrca_uncertainty_between(
            hstrat.col_to_specimen(a),
            hstrat.col_to_specimen(b),
            prior,
        ) == hstrat.calc_rank_of_mrca_uncertainty_between(a, b, prior)



import typing

import numpy as np
import pandas as pd

from ..._auxiliary_lib import (
    alifestd_find_leaf_ids,
    alifestd_has_contiguous_ids,
    alifestd_has_multiple_roots,
    alifestd_is_topologically_sorted,
    alifestd_to_working_format,
    alifestd_topological_sort,
    cast_int_lossless,
    give_len,
    unfurl_lineage_with_contiguous_ids,
)
from ..._auxiliary_lib._alifestd_assign_contiguous_ids import _reassign_ids
from ...genome_instrumentation import HereditaryStratigraphicColumn
from ._descend_template_phylogeny import descend_template_phylogeny


def descend_template_phylogeny_alifestd(
    phylogeny_df: pd.DataFrame,
    seed_column: HereditaryStratigraphicColumn,
    extant_ids: typing.Optional[typing.Iterable[int]] = None,
    progress_wrap: typing.Callable = lambda x: x,
) -> typing.List[HereditaryStratigraphicColumn]:
    """Generate a population of hereditary stratigraphic columns that could
    have resulted from the template phylogeny.

    Parameters
    ----------
    phylogeny_df : pandas.DataFrame
        Phylogeny record as an alife-standard DataFrame.
    seed_column : HereditaryStratigraphicColumn
        Hereditary stratigraphic column to seed at root node of phylogeny.

        Returned hereditary stratigraphic column population will be generated
        as if repeatedly calling `CloneDescendant()` on `seed_column`. As such,
        specifies configuration (i.e., differentia bit width and stratum
        retention policy) for returned columns. May already have strata
        deposited, which will be incorporated into generated extant population.
    extant_ids : optional list of int
        Which organisms should hereditary stratigraphic columns be created for?

        Designates content and order of returned list of hereditary
        stratigraphic column, with id values corresponding to "id" column in
        `phylogeny_df`.

        If None, hereditary stratigraphic columns will be created for all
        phylogenetic leaves (organisms without offspring) in order of
        appearance in `phylogeny_df`.
    progress_wrap : Callable, default identity function
        Wrapper applied around generation iterator and row generator for final
        phylogeny compilation process.

        Pass tqdm or equivalent to display progress bars.

    Returns
    -------
    list of HereditaryStratigraphicColumn
        Population of hereditary stratigraphic columns for extant lineage
        members (i.e., phylogeny leaf nodes).

        Columns ordered in order of appearance of corresponding extant organism
        id.
    """

    assert not alifestd_has_multiple_roots(phylogeny_df)

    # must take leaf_ids before possible topological sort to preserve order
    if extant_ids is None:
        extant_ids = alifestd_find_leaf_ids(phylogeny_df)

    working_df = alifestd_to_working_format(phylogeny_df)
    assert "ancestor_id" in working_df

    if not alifestd_has_contiguous_ids(phylogeny_df):
        if not alifestd_is_topologically_sorted(phylogeny_df):
            phylogeny_df = alifestd_topological_sort(phylogeny_df)

        __, extant_ids = _reassign_ids(
            phylogeny_df["id"].to_numpy(), np.fromiter(extant_ids, int)
        )

    ancestor_id_lookup = working_df["ancestor_id"].iloc

    def lookup_ancestor_id(id_: int) -> int:
        return ancestor_id_lookup[id_]

    def get_stem_length(id_: int) -> int:
        ancestor_id = lookup_ancestor_id(id_)
        if ancestor_id == id_:
            return 0
        elif "origin_time" in working_df:
            res = (
                working_df["origin_time"].iloc[id_]
                - working_df["origin_time"].iloc[ancestor_id]
            )
            assert res >= 0
            return cast_int_lossless(
                res, action="warn", context="origin time difference"
            )
        elif "generation" in working_df:
            res = (
                working_df["generation"].iloc[id_]
                - working_df["generation"].iloc[ancestor_id]
            )
            assert res >= 0
            return cast_int_lossless(
                res, action="warn", context="origin time difference"
            )
        else:
            return 1

    return descend_template_phylogeny(
        give_len(
            (
                unfurl_lineage_with_contiguous_ids(
                    working_df["ancestor_id"].to_numpy(),
                    leaf_id,
                )
                for leaf_id in extant_ids
            ),
            len(extant_ids),
        ),
        working_df["id"].to_numpy(),  # descending_tree_iterable
        lookup_ancestor_id,
        get_stem_length,
        seed_column,
        demark=lambda x: x,
        progress_wrap=progress_wrap,
    )



import pytest

from hstrat._auxiliary_lib import pairwise, unpairwise


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_unpairwise_empty(wrap):
    assert [*unpairwise(wrap([]))] == []


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_unpairwise_single_pair(wrap):
    assert [*unpairwise(wrap([(1, 2)]))] == [1, 2]


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_unpairwise_multiple_two_pairs(wrap):
    assert [*unpairwise(wrap([(1, 2), (3, 4)]))] == [1, 2, 4]


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_unpairwise_three_pairs(wrap):
    assert [*unpairwise(wrap([(1, 2), (3, 4), (5, 6)]))] == [1, 2, 4, 6]


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_unpairwise_four_pairs(wrap):
    assert [*unpairwise(wrap([(1, 2), (3, 4), (5, 6), (7, 8)]))] == [
        1,
        2,
        4,
        6,
        8,
    ]


@pytest.mark.parametrize("wrap", [iter, list, lambda x: x])
@pytest.mark.parametrize("size", range(20))
def test_unpairwise_inverse_pairwise(wrap, size):
    a = range(size)
    if size == 1:
        assert [*unpairwise(wrap(pairwise(a)))] == []
    else:
        assert [*unpairwise(wrap(pairwise(a)))] == [*a]



from copy import copy
import itertools as it
import typing

import anytree

from ..._auxiliary_lib import AnyTreeAscendingIter
from .._HereditaryStratum import HereditaryStratum
from ._detail import HereditaryStratumOrderedStoreBase


class HereditaryStratumOrderedStoreTree(HereditaryStratumOrderedStoreBase):
    """Interchangeable backing container for HereditaryStratigraphicColumn.

    Stores deposited strata as a linked tree. Retained strata pertinent to the
    store are essentially a linear linked list reaching from tree leaf (most
    recent stratum) back to the tree root (most ancient stratum). Cloned stores
    directly share common components of the tree and then branch out different
    leaves as subsequent strata are deposited. Stratum deletion results in the
    entire strata sequence more recent than the deleted stratum being copied
    into an independent branch for the store that requested the deletion.

    Potentially useful in scenarios where stratum deletions are uncommon (i.e.,
    the perfect resolution stratum retention policy) or column cloning occurs heavily without much stratum deposition.
    """

    __slots__ = (
        "_leaf",
        "_num_strata_retained",
    )

    # strata stored in a tree with most ancient as root and most recent as leaf
    _leaf: anytree.AnyNode  # will contain HereditaryStratum
    # maintaining a counter is much more efficient than counting steps from leaf
    # to root
    _num_strata_retained: int

    def __init__(self: "HereditaryStratumOrderedStoreTree"):
        """Initialize instance variables."""
        self._leaf = None
        self._num_strata_retained = 0

    def __del__(self: "HereditaryStratumOrderedStoreTree"):
        """Destruct the store.

        Releases resources no longer needed by this store that are needed by no
        other stores.

        anytree uses bidirectional references between nodes so manual pruning
        is necessary because garbage collection won't recognize unused nodes as
        fully unreachable.
        """
        # delete all nodes that only lead to this store's leaf
        for node in self._GetAscendingIter():
            if sum(1 for _ in zip(node.children, range(2))) == 2:
                # if node has more than one child, stop deleting
                break
            else:
                # delete and keep ascending
                node.parent = None
                del node

    def __eq__(
        self: "HereditaryStratumOrderedStoreTree",
        other: "HereditaryStratumOrderedStoreTree",
    ) -> bool:
        """Compare for value-wise equality."""
        return isinstance(other, self.__class__) and (
            self._leaf.stratum if self._leaf is not None else None,
            self._num_strata_retained,
        ) == (
            other._leaf.stratum if other._leaf is not None else None,
            other._num_strata_retained,
        )

    def _GetAscendingIter(
        self: "HereditaryStratumOrderedStoreTree",
    ) -> typing.Iterator[anytree.AnyNode]:
        """Iterate over retained strata from most recent to most ancient."""
        return AnyTreeAscendingIter(self._leaf)

    def DepositStratum(
        self: "HereditaryStratumOrderedStoreTree",
        rank: typing.Optional[int],
        stratum: "HereditaryStratum",
    ) -> None:
        """Insert a new stratum into the store.

        Parameters
        ----------
        rank : typing.Optional[int]
            The position of the stratum being deposited within the sequence of
            strata deposited into the column. Precisely, the number of strata
            that have been deposited before stratum.
        stratum : HereditaryStratum
            The stratum to deposit.
        """
        new_leaf = anytree.AnyNode(stratum=stratum)
        new_leaf.parent = self._leaf
        self._leaf = new_leaf
        self._num_strata_retained += 1

    def GetNumStrataRetained(self: "HereditaryStratumOrderedStoreTree") -> int:
        """How many strata are present in the store?

        May be fewer than the number of strata deposited if deletions have
        occured.
        """
        return self._num_strata_retained

    def GetStratumAtColumnIndex(
        self: "HereditaryStratumOrderedStoreTree",
        index: int,
        # needed for other implementations
        get_rank_at_column_index: typing.Optional[typing.Callable] = None,
    ) -> HereditaryStratum:
        """Get the stratum positioned at index i among retained strata.

        Index order is from most ancient (index 0) to most recent.

        Parameters
        ----------
        ranks : iterator over int
            The ranks that are to be deleted.
        get_column_index_of_rank : callable, optional
            Callable that returns the index position within retained strata of
            the stratum deposited at rank r.
        """
        return next(
            it.islice(
                self._GetAscendingIter(),
                self.GetNumStrataRetained() - 1 - index,
                None,
            )
        ).stratum

    def GetRankAtColumnIndex(
        self: "HereditaryStratumOrderedStoreTree",
        index: int,
    ) -> int:
        """Map from deposition generation to column position.

        What is the deposition rank of the stratum positioned at index i
        among retained strata?
        """
        res_rank = self.GetStratumAtColumnIndex(index).GetDepositionRank()
        assert res_rank is not None
        return res_rank

    def GetColumnIndexOfRank(
        self: "HereditaryStratumOrderedStoreTree",
        rank: int,
    ) -> typing.Optional[int]:
        """Map from column position to deposition generation.

        What is the index position within retained strata of the stratum
        deposited at rank r? Returns None if no stratum with rank r is present
        within the store.
        """
        try:
            return next(
                self.GetNumStrataRetained() - 1 - idx
                for idx, node in enumerate(self._GetAscendingIter())
                if node.stratum.GetDepositionRank() == rank
            )
        except StopIteration:
            return None

    def _do_getrank_DelRanks(
        self: "HereditaryStratumOrderedStoreTree",
        ranks: typing.Iterator[int],
        # deposition ranks are stored in strata
    ) -> None:
        """Provide implementation detail for DelRanks.

        Handles case where deposition ranks are stored in strata.
        """
        # duplicate everything after deepest deletion
        # except other ranks slated for deletion
        pending_node = (
            None  # top of new chain, needs to be attached to a parent
        )

        ascending_iter = self._GetAscendingIter()
        target_ranks = sorted(ranks)
        # pop ranks off back of target_ranks as they're deleted
        while target_ranks:
            cur_node = next(ascending_iter)

            if cur_node.stratum.GetDepositionRank() == target_ranks[-1]:
                # skip over and don't copy node targeted for deletion
                target_ranks.pop()
                self._num_strata_retained -= 1
            else:
                # copy untargeted node and use it as parent for pending node
                new_node = anytree.AnyNode(
                    stratum=cur_node.stratum,
                )
                if pending_node is None:
                    self._leaf = new_node
                else:
                    pending_node.parent = new_node
                pending_node = new_node

        # hook copied content into existing tree
        if pending_node is not None:
            pending_node.parent = next(ascending_iter, None)
        else:
            self._leaf = next(ascending_iter, None)

    def _do_calcrank_DelRanks(
        self: "HereditaryStratumOrderedStoreTree",
        ranks: typing.Iterator[int],
        # deposition ranks are not stored in strata
        get_column_index_of_rank: typing.Callable,
    ) -> None:
        """Provide implementation detail for DelRanks.

        Handles case where deposition ranks are not stored in strata but a
        function to calculate deposition rank from column index is available.
        """
        # duplicate everything after deepest deletion
        # except other ranks slated for deletion

        pending_node = (
            None  # top of new chain, needs to be attached to a parent
        )
        pending_leaf = None  # bottom of new chain
        num_deleted_nodes = 0
        # RE: pending_leaf, num_deleted_nodes
        # get_column_index_of_rank can depend on state of this object
        # so we have to cache changes and apply all at once at the end

        ascending_iter = self._GetAscendingIter()
        ascending_idx = self.GetNumStrataRetained() - 1

        for rank in sorted(ranks, reverse=True):
            target_idx = get_column_index_of_rank(rank)
            assert ascending_idx >= target_idx

            while ascending_idx >= target_idx:
                cur_node = next(ascending_iter)

                if ascending_idx == target_idx:
                    # don't copy node targeted for deletion
                    num_deleted_nodes += 1
                else:
                    # copy untargeted node and use it as parent for pending node
                    new_node = anytree.AnyNode(
                        stratum=cur_node.stratum,
                    )
                    if pending_node is None:
                        pending_leaf = new_node
                    else:
                        pending_node.parent = new_node
                    pending_node = new_node

                # step backward
                ascending_idx -= 1

            assert ascending_idx == target_idx - 1

        # hook copied content into existing tree
        if pending_node is not None:
            pending_node.parent = next(ascending_iter, None)
        else:
            assert pending_leaf is None
            pending_leaf = next(ascending_iter, None)
        # attach self to new copied chain
        if pending_leaf is not None:
            self._leaf = pending_leaf
        # apply cached changes to stratum count
        self._num_strata_retained -= num_deleted_nodes

    def DelRanks(
        self: "HereditaryStratumOrderedStoreTree",
        ranks: typing.Iterator[int],
        # deposition ranks might not be stored in strata
        get_column_index_of_rank: typing.Optional[typing.Callable] = None,
    ) -> None:
        """Purge strata with specified deposition ranks from the store.

        Parameters
        ----------
        ranks : iterator over int
            The ranks that are to be deleted.
        get_column_index_of_rank : callable, optional
            Callable that returns the deposition rank of the stratum positioned at index i among retained strata.
        """
        if get_column_index_of_rank is None:
            self._do_getrank_DelRanks(ranks)
        else:
            self._do_calcrank_DelRanks(ranks, get_column_index_of_rank)

    def IterRetainedRanks(
        self: "HereditaryStratumOrderedStoreTree",
    ) -> typing.Iterator[int]:
        """Iterate over deposition ranks of strata present in the store from
        most ancient to most recent.

        The store may be altered during iteration without iterator
        invalidation, although subsequent updates will not be reflected in the
        iterator.
        """
        # must make copy to prevent invalidation when strata are deleted
        # note, however, that copy is made lazily
        # (only when first item requested)
        ranks = [
            node.stratum.GetDepositionRank()
            for node in self._GetAscendingIter()
        ]
        for rank in reversed(ranks):
            assert rank is not None
            yield rank

    def IterRetainedStrata(
        self: "HereditaryStratumOrderedStoreTree",
    ) -> typing.Iterator[HereditaryStratum]:
        """Iterate over stored strata from most ancient to most recent."""
        yield from reversed(
            [node.stratum for node in self._GetAscendingIter()]
        )

    def _do_reverse_IterRankDifferentiaZip(
        self: "HereditaryStratumOrderedStoreTree",
        # deposition ranks might not be stored in strata
        get_rank_at_column_index: typing.Optional[typing.Callable] = None,
        start_column_index: int = 0,
    ) -> typing.Iterator[typing.Tuple[int, int]]:
        """Iterate over differentia and corresponding depsotion rank.

        Ordered from most recent to most ancient. Implementation detail for
        IterRankDifferentiaZip.
        """
        for reverse_column_idx, node in enumerate(self._GetAscendingIter()):
            column_idx = self.GetNumStrataRetained() - 1 - reverse_column_idx
            if column_idx >= start_column_index:
                rank: int
                if get_rank_at_column_index is None:
                    rank = node.stratum.GetDepositionRank()
                    assert rank is not None
                else:
                    rank = get_rank_at_column_index(column_idx)
                    assert rank is not None
                yield (rank, node.stratum.GetDifferentia())
            else:
                break

    def IterRankDifferentiaZip(
        self: "HereditaryStratumOrderedStoreTree",
        # deposition ranks might not be stored in strata
        get_rank_at_column_index: typing.Optional[typing.Callable] = None,
        start_column_index: int = 0,
    ) -> typing.Iterator[typing.Tuple[int, int]]:
        """Iterate over differentia and corresponding depsotion rank.

        Values yielded as tuples. Guaranteed ordered from most ancient to most
        recent.

        Parameters
        ----------
        get_rank_at_column_index : callable, optional
            Callable that returns the deposition rank of the stratum
            positioned at index i among retained strata.
        start_column_index : callable, optional
            Number of strata to skip over before yielding first result from the
            iterator. Default 0, meaning no strata are skipped over.
        """
        reverse_iter = self._do_reverse_IterRankDifferentiaZip(
            get_rank_at_column_index=get_rank_at_column_index,
            start_column_index=start_column_index,
        )
        reverse_data = [*reverse_iter]
        return reversed(reverse_data)

    def Clone(
        self: "HereditaryStratumOrderedStoreTree",
    ) -> "HereditaryStratumOrderedStoreTree":
        """Create an independent copy of the store.

        Returned copy contains identical data but may be freely altered without
        affecting data within this store.
        """
        # shallow copy
        res = copy(self)
        # must create independent leaf
        if self._leaf is not None:
            res._leaf = anytree.AnyNode(
                stratum=self._leaf.stratum,
            )
            res._leaf.parent = self._leaf.parent
        return res



"""Implementation for `evolve_fitness_trait_population`."""

from ._apply_island_migrations import _apply_island_migrations
from ._apply_mutation import _apply_mutation
from ._apply_niche_invasions import _apply_niche_invasions
from ._get_island_id import _get_island_id
from ._get_niche_id import _get_niche_id
from ._select_parents import _select_parents

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "_apply_island_migrations",
    "_apply_mutation",
    "_apply_niche_invasions",
    "_get_island_id",
    "_get_niche_id",
    "_select_parents",
]



import itertools as it
import typing

import pytest

from hstrat._auxiliary_lib import pairwise, popcount
from hstrat.hstrat import recency_proportional_resolution_curbed_algo


def _iter_backing_policy_transition_ranks(
    size_curb: int,
) -> typing.Iterator[int]:

    geom_seq_nth_root_algo_transition_rank = 2 ** (size_curb - 1) + 1
    for i in it.count():
        cur_rank = 2**i
        assert popcount(cur_rank) == 1

        if cur_rank >= geom_seq_nth_root_algo_transition_rank:
            return geom_seq_nth_root_algo_transition_rank

        yield cur_rank


@pytest.mark.parametrize(
    "size_curb",
    [
        1,
        2,
        3,
        7,
        8,
        9,
        42,
        97,
        100,
    ],
)
def test_implementation_consistency_at_backing_policy_transitions(size_curb):
    policy = recency_proportional_resolution_curbed_algo.Policy(size_curb)
    spec = policy.GetSpec()
    instance = recency_proportional_resolution_curbed_algo.IterRetainedRanks(
        spec
    )

    for num_strata_deposited in (
        test_rank
        for transition_rank in _iter_backing_policy_transition_ranks(size_curb)
        for test_rank in range(
            max(transition_rank - 2, 0), transition_rank + 3
        )
    ):

        for which1, which2 in pairwise(
            (
                instance,
                recency_proportional_resolution_curbed_algo.IterRetainedRanks(
                    spec
                ),
            )
        ):
            # fmt: off
            # black bug in 2022.{10,12}.0 misformats {*x}
            assert {
                *which1(
                    policy,
                    num_strata_deposited,
                )
            } == {
                *which2(
                    policy,
                    num_strata_deposited,
                )
            }
            # fmt: on



import math

from ._cmp import cmp


def cmp_approx(a: float, b: float) -> int:
    """
    Compare two floating-point numbers and return an integer indicating their
    relative order, considering close values equal.

    Parameters
    ----------
    a : float
        The first number to compare.
    b : float
        The second number to compare.

    Returns
    -------
    int
        1 if `a` is greater than `b`, -1 if `a` is less than `b`, 0 if `a` and
        `b` are close.

    Example
    -------
    >>> cmp_approx(1.0, 1.00000001)
    0
    >>> cmp_approx(2.0, 1.0)
    1
    >>> cmp_approx(1.0, 2.0)
    -1
    >>> cmp_approx(2.0, 2.0)
    0
    """
    return 0 if math.isclose(a, b) else cmp(a, b)



import typing

from ._RngStateContext import RngStateContext


def with_rng_state_context(seed: int) -> typing.Callable:
    """Decorator to set the random number generator state before calling a
    function.

    The pre-existing random number generator state is restored when the
    decorated function returns.

    Parameters
    ----------
    seed : int
        The seed to use for the random number generator.

    Returns
    -------
    typing.Callable
        A callable that takes a function as its argument and returns a
        decorated function.

    Examples
    --------
    >>> import random
    >>> @with_rng_state_context(seed=123)
    ... def my_function():
    ...     return random.random()
    ...
    >>> my_function() == my_function()
    True
    """

    def decorator(func: typing.Callable) -> typing.Callable:
        """Decorate a function to set the random number generator state.

        Parameters
        ----------
        func : typing.Callable
            The function to be decorated.

        Returns
        -------
        typing.Callable
            The decorated function.
        """

        def wrapper(*args, **kwargs):
            """Wrapper function that sets the random number generator state
            before calling the decorated function.

            Parameters
            ----------
            args : list
                The positional arguments to pass to the decorated function.
            kwargs : dict
                The keyword arguments to pass to the decorated function.

            Returns
            -------
            The return value of the decorated function.
            """
            with RngStateContext(seed):
                return func(*args, **kwargs)

        return wrapper

    return decorator



import anytree

from hstrat._auxiliary_lib import anytree_has_sibling


def test_no_sibling_lone_root():
    # Create a tree with just one node
    root = anytree.Node("root")
    # Assert that the node has no siblings
    assert not anytree_has_sibling(root)


def test_no_sibling_only_child():
    # Create a tree with just one node
    root = anytree.Node("root")
    child1 = anytree.Node("child1", parent=root)
    # Assert that the node has no siblings
    assert not anytree_has_sibling(child1)


def test_single_sibling():
    # Create a tree with two sibling nodes
    root = anytree.Node("root")
    child1 = anytree.Node("child1", parent=root)
    child2 = anytree.Node("child2", parent=root)
    # Assert that child1 has at least one sibling
    assert anytree_has_sibling(child1)
    # Assert that child2 has at least one sibling
    assert anytree_has_sibling(child2)


def test_multiple_siblings():
    # Create a tree with multiple sibling nodes
    root = anytree.Node("root")
    child1 = anytree.Node("child1", parent=root)
    anytree.Node("child2", parent=root)
    grandchild1 = anytree.Node("grandchild1", parent=child1)
    grandchild2 = anytree.Node("grandchild2", parent=child1)
    # Assert that grandchild1 has at least one sibling
    assert anytree_has_sibling(grandchild1)
    # Assert that grandchild2 has at least one sibling
    assert anytree_has_sibling(grandchild2)
    # Assert that root has no siblings
    assert not anytree_has_sibling(root)


def test_root_node():
    # Create a tree with multiple nodes
    root = anytree.Node("root")
    anytree.Node("child1", parent=root)
    anytree.Node("child2", parent=root)
    # Assert that the root node has no siblings
    assert not anytree_has_sibling(root)


def test_leaf_node():
    # Create a tree with multiple nodes
    root = anytree.Node("root")
    child1 = anytree.Node("child1", parent=root)
    anytree.Node("child2", parent=root)
    anytree.Node("grandchild1", parent=child1)
    grandchild2 = anytree.Node("grandchild2", parent=child1)
    # Assert that the leaf node grandchild2 has at least one sibling
    assert anytree_has_sibling(grandchild2)



from iterpop import iterpop as ip
import pandas as pd

from hstrat._auxiliary_lib import alifestd_find_root_ids
import hstrat.phylogenetic_inference.tree._impl as impl


def test_time_calibrate_tree_with_simple_tree():

    alifestd_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4],
            "ancestor_list": ["[1]", "[none]", "[1]", "[2]", "[2]"],
            "branch_length": [8, 0, 2, 1, 11],
            "name": ["A", "B", "C", "D", "E"],
        }
    )
    leaf_node_origin_times = {3: 52, 4: 62}  # 3 -> "D", 4 -> "E"

    for df in alifestd_df, alifestd_df.sample(frac=1):
        df_ = df.copy()
        calibrated_df = impl.time_calibrate_tree(df, leaf_node_origin_times)

        assert df.equals(df_)
        assert ip.popsingleton(alifestd_find_root_ids(calibrated_df)) == 0
        assert dict(
            zip(calibrated_df["name"], calibrated_df["origin_time"])
        ) == {
            "E": 62,
            "D": 52,
            "C": 51,
            "B": 49,
            "A": 41,
        }


def test_time_calibrate_tree_with_empty_tree():
    alifestd_df = pd.DataFrame(
        {
            "id": [],
            "ancestor_list": [],
            "branch_length": [],
            "name": [],
        }
    )
    alifestd_df_ = alifestd_df.copy()
    leaf_node_origin_times = {}
    calibrated_df = impl.time_calibrate_tree(
        alifestd_df, leaf_node_origin_times
    )
    assert calibrated_df.equals(alifestd_df_)
    assert alifestd_df.equals(alifestd_df_)


def test_time_calibrate_tree_with_singleton():
    alifestd_df = pd.DataFrame(
        {
            "id": [0],
            "ancestor_list": ["[none]"],
            "branch_length": [0],
            "name": ["Q"],
        }
    )
    leaf_node_origin_times = {0: 42}  # 0 -> "Q"

    expected_df = pd.DataFrame(
        {
            "id": [0],
            "ancestor_list": ["[none]"],
            "ancestor_id": [0],
            "origin_time": [42],
            "name": ["Q"],
        }
    )

    alifestd_df_ = alifestd_df.copy()
    calibrated_df = impl.time_calibrate_tree(
        alifestd_df, leaf_node_origin_times
    )
    assert calibrated_df.sort_index(axis="columns").equals(
        expected_df.sort_index(axis="columns")
    )
    assert alifestd_df.equals(alifestd_df_)


def test_time_calibrate_tree_with_disjoint_tree1():
    alifestd_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_list": ["[1]", "[none]", "[1]", "[2]", "[2]", "[none]"],
            "branch_length": [8, 0, 2, 1, 11, 99],
            "name": ["A", "B", "C", "D", "E", "Q"],
        }
    )
    leaf_node_origin_times = {3: 52, 4: 62, 5: 11}  # 3->D, 4->E, 5->Q

    for df in alifestd_df, alifestd_df.sample(frac=1):
        df_ = df.copy()
        calibrated_df = impl.time_calibrate_tree(df, leaf_node_origin_times)

        assert df.equals(df_)
        assert set(alifestd_find_root_ids(calibrated_df)) == {0, 5}
        assert dict(
            zip(calibrated_df["name"], calibrated_df["origin_time"])
        ) == {
            "E": 62,
            "D": 52,
            "C": 51,
            "B": 49,
            "A": 41,
            "Q": 11,
        }


def test_time_calibrate_tree_with_disjoint_tree2():
    alifestd_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[1]",
                "[none]",
                "[1]",
                "[2]",
                "[2]",
                "[none]",
                "[5]",
            ],
            "branch_length": [8, 0, 2, 1, 11, 0, 1],
            "name": ["A", "B", "C", "D", "E", "Q", "R"],
        }
    )
    leaf_node_origin_times = {3: 52, 4: 62, 6: 1}  # 3->D, 4->E, 6->R,

    for df in alifestd_df, alifestd_df.sample(frac=1):
        df_ = df.copy()
        calibrated_df = impl.time_calibrate_tree(df, leaf_node_origin_times)

        assert df.equals(df_)
        assert set(alifestd_find_root_ids(calibrated_df)) == {0, 5}
        assert dict(
            zip(calibrated_df["name"], calibrated_df["origin_time"])
        ) == {
            "E": 62,
            "D": 52,
            "C": 51,
            "B": 49,
            "A": 41,
            "Q": 0,
            "R": 1,
        }


def test_time_calibrate_tree_with_disjoint_tree3():
    alifestd_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[1]",
                "[none]",
                "[1]",
                "[2]",
                "[2]",
                "[6]",
                "[none]",
            ],
            "branch_length": [8, 0, 2, 1, 11, 1, 0],
            "name": ["A", "B", "C", "D", "E", "Q", "R"],
        }
    )
    leaf_node_origin_times = {3: 52, 4: 62, 6: 1}  # 3->D, 4->E, 6->R,

    for df in alifestd_df, alifestd_df.sample(frac=1):
        df_ = df.copy()
        calibrated_df = impl.time_calibrate_tree(df, leaf_node_origin_times)

        assert df.equals(df_)
        assert set(alifestd_find_root_ids(calibrated_df)) == {0, 5}
        assert dict(
            zip(calibrated_df["name"], calibrated_df["origin_time"])
        ) == {
            "E": 62,
            "D": 52,
            "C": 51,
            "B": 49,
            "A": 41,
            "Q": 0,
            "R": 1,
        }


def test_time_calibrate_tree_with_zero_length_branch_tree():

    alifestd_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4],
            "ancestor_list": ["[1]", "[none]", "[1]", "[2]", "[2]"],
            "branch_length": [8, 0, 2, 1, 0],
            "name": ["A", "B", "C", "D", "E"],
        }
    )
    leaf_node_origin_times = {3: 52, 4: 51}  # 3 -> "D", 4 -> "E"

    for df in alifestd_df, alifestd_df.sample(frac=1):
        df_ = df.copy()
        calibrated_df = impl.time_calibrate_tree(df, leaf_node_origin_times)

        assert df.equals(df_)
        assert ip.popsingleton(alifestd_find_root_ids(calibrated_df)) == 0
        assert dict(
            zip(calibrated_df["name"], calibrated_df["origin_time"])
        ) == {
            "E": 51,
            "D": 52,
            "C": 51,
            "B": 49,
            "A": 41,
        }



import unittest

from hstrat._auxiliary_lib import is_strictly_decreasing


class TestIsStrictlyDecreasing(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_empty(self):
        assert is_strictly_decreasing([])

    def test_singleton(self):
        assert is_strictly_decreasing(["a"])
        assert is_strictly_decreasing([0])
        assert is_strictly_decreasing([1])

    def test_nondecreasing(self):
        assert is_strictly_decreasing(
            reversed(
                [
                    *range(10),
                ]
            )
        )
        assert not is_strictly_decreasing(
            reversed(
                [
                    0,
                    *range(10),
                ]
            )
        )
        assert not is_strictly_decreasing(
            reversed(
                [
                    0,
                    0,
                    *range(10),
                    *range(9, 18),
                ]
            )
        )

    def test_decreasing(self):
        assert not is_strictly_decreasing(
            [
                -1,
                0,
            ]
        )
        assert not is_strictly_decreasing(
            reversed(
                [
                    *range(10),
                    *range(2),
                ]
            )
        )


if __name__ == "__main__":
    unittest.main()



from ._IterRetainedRanksProduction import IterRetainedRanksProduction
from ._IterRetainedRanksReference import IterRetainedRanksReference

impls = (IterRetainedRanksProduction, IterRetainedRanksReference)



import colorsys
import typing

import matplotlib as mpl


# adapted from https://stackoverflow.com/a/49601444
def scale_luminosity(
    color: typing.Union[typing.Tuple[float, float, float], str],
    amount: float,
) -> typing.Tuple[float, float, float]:
    """Lighten or darken a given color.

    Multiplies luminosity by the given amount. Input can be matplotlib color
    string, hex string, or RGB tuple.

    Examples
    --------
    >> scale_luminosity('g', 0.3)
    >> scale_luminosity('#F034A3', 0.6)
    >> scale_luminosity((.3,.55,.1), 0.5)
    """
    try:
        c = mpl.colors.cnames[color]
    except KeyError:
        c = color
    c = colorsys.rgb_to_hls(*mpl.colors.to_rgb(c))
    return colorsys.hls_to_rgb(c[0], max(0, min(1, amount * c[1])), c[2])



import pandas as pd


def alifestd_is_sexual(phylogeny_df: pd.DataFrame) -> bool:
    """Do any organisms in the phylogeny have than one immediate ancestor?

    Input dataframe is not mutated by this operation.
    """
    return phylogeny_df["ancestor_list"].astype("str").str.contains(",").any()



from numbers import Number

import numpy as np
import pytest

from hstrat._auxiliary_lib import demark


@pytest.mark.parametrize(
    "collection",
    [
        range(20),
        "abcdefghij",
        np.arange(10),
        np.linspace(1, 42),
        np.arange(20, dtype=float),
        [{x} for x in range(20)],
        [{3} for x in range(20)],
    ],
)
def test_demark(collection):
    assert len(collection) == len(set(map(demark, collection)))
    for x in collection:
        assert demark(x) == demark(x)
        if isinstance(x, (int, float, Number, str)):
            assert demark(x) == x



import typing


# adapted from https://stackoverflow.com/a/434328
def iter_chunks(
    seq: typing.Sequence,
    chunk_size: int,
    start: int = 0,
) -> typing.Iterator[typing.Sequence]:
    """Yield successive chunks of size `chunk_size` from sequence `seq`.

    Parameters
    ----------
    seq : Sequence
        The sequence to be chunked.
    chunk_size : int
        The size of each chunk.

    Returns
    -------
    Iterator[Sequence]
        A generator object that yields chunks of `seq` of size `chunk_size`.

    Examples
    --------
    >>> seq = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    >>> chunk_size = 3
    >>> for chunk in iter_chunks(seq, chunk_size):
    ...     print(chunk)
    [1, 2, 3]
    [4, 5, 6]
    [7, 8, 9]
    [10]
    """
    return (
        seq[pos : pos + chunk_size]
        for pos in range(start, len(seq), chunk_size)
    )



import itertools as it
import typing

import numpy as np

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ..pairwise import estimate_patristic_distance_between


def build_distance_matrix_numpy(
    population: typing.Sequence[HereditaryStratigraphicArtifact],
    estimator: str,
    prior: typing.Union[str, typing.Any],
    force_common_ancestry: typing.Optional[bool] = False,
) -> np.ndarray:
    """Compute all-pairs patristic distance among a population of extant
    hereditary stratigraphic columns.

    Parameters
    ----------
    population : Sequence[HereditaryStratigraphicArtifact]
        The extant hereditary stratigraphic columns to compare.

        The ordering of rows and columns within the returned matrix will
        correspond to the ordering of columns in this sequence.
    estimator : {"maximum_likelihood", "unbiased"}
        What patristic distance estimation method should be used? Options are
        "maximum_likelihood" or "unbiased".

        See `estimate_ranks_since_mrca_with` for discussion of estimator
        options.
    prior :
        Prior probability density distribution over possible generations of the
        MRCA for MRCA estimation.

        See `estimate_ranks_since_mrca_with` for discussion of prior
        options.
    force_common_ancestry : Optional[bool], optional
        How should columns that definitively share no common ancestry be handled?

        If set to True, treat columns with no common ancestry as if they
        shared a common ancestor immediately before the genesis of the
        lineages. If set to False, set the patristic distance between columns
        with no common ancestry as NaN.

        If set to None, as default, the presence of columns that definitively
        share no common ancestry will raise a ValueError.

    Returns
    -------
    np.ndarray
        The patristic distance matrix of the population, a square numpy array
        of shape (len(population), len(population)).

    Raises
    ------
    ValueError
        If the distance between two columns that definitively share no common
        ancestry are attempted to be computed without setting
        `force_common_ancestry`.
    """

    matrix = np.empty((len(population), len(population)))

    pairings = it.combinations_with_replacement(range(len(population)), 2)

    for a, b in pairings:
        patristic_distance = estimate_patristic_distance_between(
            population[a],
            population[b],
            estimator,
            prior,
        )

        if patristic_distance is None and force_common_ancestry is None:
            raise ValueError(
                f"Definitive independent origins detected for {a} and {b};"
                "Set force_common_ancestry to handle this case."
            )
        elif patristic_distance is None and force_common_ancestry:
            max_patristic_distance = (
                population[a].GetNumStrataDeposited()
                + population[b].GetNumStrataDeposited()
            )
            matrix[a][b] = max_patristic_distance
            matrix[b][a] = max_patristic_distance
        elif patristic_distance is None and not force_common_ancestry:
            matrix[a][b] = np.nan
            matrix[b][a] = np.nan
        elif a == b:
            # this is important to do manually:
            # estimators are not guaranteed to yield zero or near zero
            # patristic distance for identical columns
            matrix[a][b] = 0.0
            matrix[b][a] = 0.0
        else:
            matrix[a][b] = patristic_distance
            matrix[b][a] = patristic_distance

    return matrix



import contextlib
import logging
import time
import typing


@contextlib.contextmanager
def log_context_duration(
    what: str,
    logger: callable = logging.info,
) -> typing.Iterable[None]:
    start = time.time()
    yield
    elapsed = time.time() - start
    logger(
        f"""exit log_seconds_elapsed for {what}
!!! {{"{what}": {elapsed}}}""",
    )



import typing

import matplotlib as mpl
from matplotlib import pyplot as plt
import seaborn as sns

from ...genome_instrumentation._HereditaryStratigraphicColumn import (
    HereditaryStratigraphicColumn,
)


def strata_retained_num_lineplot(
    stratum_retention_policy: typing.Callable[[int, int], bool],
    num_generations: int,
    do_show: bool = False,
    ax: typing.Optional[plt.matplotlib.axes.Axes] = None,
) -> plt.matplotlib.axes.Axes:
    """Plot number deposited strata that are retained at each generation.

    Parameters
    ----------
    stratum_retention_policy : any
        Object specifying stratum retention policy.
    num_generations : int
        Number of generations to plot.
    ax : matplotlib/pylab axes, optional
        If a valid matplotlib.axes.Axes instance, the plot is drawn in that
        Axes. By default (None), a new axes is created.
    do_show : bool, optional
        Whether to show() the plot automatically.
    """
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
    elif not isinstance(ax, plt.matplotlib.axes.Axes):
        raise ValueError(f"Invalid argument for ax: {ax}")

    xs = [0]
    ys = [0]
    column = HereditaryStratigraphicColumn(
        stratum_retention_policy=stratum_retention_policy,
    )
    for gen in range(1, num_generations):
        xs.append(gen)
        if stratum_retention_policy.CalcNumStrataRetainedExact is not None:
            ys.append(
                stratum_retention_policy.CalcNumStrataRetainedExact(gen),
            )
        else:
            ys.append(column.GetNumStrataRetained())
            column.DepositStratum()

    sns.lineplot(
        x=xs,
        y=ys,
        ax=ax,
    )
    ax.set_xlabel("Generation")
    ax.set_ylabel("Num Strata Retained")

    ax.xaxis.set_major_locator(
        mpl.ticker.MaxNLocator(
            nbins="auto",
            steps=[1, 2, 5, 10],
            integer=True,
            min_n_ticks=0,
        )
    )
    ax.yaxis.set_major_locator(
        mpl.ticker.MaxNLocator(
            nbins="auto",
            steps=[1, 2, 5, 10],
            integer=True,
            min_n_ticks=0,
        )
    )
    ymin, ymax = ax.get_ylim()
    ax.set_ylim([ymin, ymax + 2])

    # strip any negative xticks
    ax.set_xticks([tick for tick in ax.get_xticks() if tick >= 0])

    if do_show:
        plt.show()

    return ax



"""
Must be defined in a separate file so that it is able to be imported
by the attach_stub function and loaded into the init. The module of this
list is laundered making it essentially defined therein.
"""

from . import (
    depth_proportional_resolution_algo,
    depth_proportional_resolution_tapered_algo,
    fixed_resolution_algo,
    geom_seq_nth_root_algo,
    geom_seq_nth_root_tapered_algo,
    nominal_resolution_algo,
    perfect_resolution_algo,
    pseudostochastic_algo,
    recency_proportional_resolution_algo,
    recency_proportional_resolution_curbed_algo,
    stochastic_algo,
)

provided_stratum_retention_algorithms = [
    depth_proportional_resolution_algo,
    depth_proportional_resolution_tapered_algo,
    fixed_resolution_algo,
    geom_seq_nth_root_algo,
    geom_seq_nth_root_tapered_algo,
    nominal_resolution_algo,
    perfect_resolution_algo,
    pseudostochastic_algo,
    recency_proportional_resolution_algo,
    recency_proportional_resolution_curbed_algo,
    stochastic_algo,
]



import random


def decide_if_discard(
    stratum_rank: int,
    num_stratum_depositions_completed: int,
    retention_probability: float,
) -> bool:
    return not (random.random() < retention_probability)
